import asyncio
from lmql.runtime.multi_head_interpretation import InterpreterReturnValue, InterpreterCall, InterpretationHead, ContinueOnly

async def fun2(context):
    result = []
    a = yield context.query("a")
    result.append(a)
    b = yield context.query("b")
    result.append(b)
    c = yield context.query("c")
    result.append(c)

    yield InterpreterReturnValue(result)

class Context:
    def query(self, value):
        return InterpreterCall(value)

async def assert_remaining_trace(head, expected, message):
    if head.waiting:
        value = await head.continue_()
    else:
        if head.iterator_fct is None:
            await head.materialize_copy_if_necessary()
        value = head.last_call_args
        if value is None: value = [None]

    if head.done:
        trace = []
    else:
        trace = [value[0]]

        while not head.done:
            await head.advance(value)
            value = await head.continue_()
            if not head.done:
                trace.append(value[0])

    assert str(trace) == str(expected), f"{message}. Expected {expected}, got {trace}"

async def test_multi_head():
    ctx = Context()
    
    # simple head
    start_head = InterpretationHead(fun2, ctx)
    await assert_remaining_trace(start_head, ["a", "b", "c"], "simple head")

    # continued head
    continued_head = InterpretationHead(fun2, ctx)
    await continued_head.continue_()
    await assert_remaining_trace(continued_head, ["a", "b", "c"], "continued head")

    # copy start_head
    start_head = InterpretationHead(fun2, ctx)
    copy_start_head = start_head.copy()
    await assert_remaining_trace(copy_start_head, ["a", "b", "c"], "copied start_head")

    # copy continued_head
    continued_head = InterpretationHead(fun2, ctx)
    await continued_head.continue_()
    copy_continued_head = continued_head.copy()
    await assert_remaining_trace(copy_continued_head, ["a", "b", "c"], "copied continued_head")

    # double copy simple head
    start_head = InterpretationHead(fun2, ctx)
    copy_start_head = start_head.copy()
    copy_copy_start_head = copy_start_head.copy()
    await assert_remaining_trace(copy_copy_start_head, ["a", "b", "c"], "double copied start_head")

    # double copy continued head
    continued_head = InterpretationHead(fun2, ctx)
    await continued_head.continue_()
    copy_continued_head = continued_head.copy()
    copy_copy_continued_head = copy_continued_head.copy()
    await assert_remaining_trace(copy_copy_continued_head, ["a", "b", "c"], "double copied continued_head")

    # advanced head
    start_head = InterpretationHead(fun2, ctx)
    value = await start_head.continue_()
    await start_head.advance(value)
    assert(start_head._internal_trace == [("a",)])
    await assert_remaining_trace(start_head, ["b", "c"], "advanced head")

    # copy advanced head
    start_head = InterpretationHead(fun2, ctx)
    value = await start_head.continue_()
    await start_head.advance(value)
    copy_start_head = start_head.copy()
    assert(copy_start_head.trace == [("a",)])
    assert(copy_start_head._internal_trace == [])
    await assert_remaining_trace(copy_start_head, ["b", "c"], "copied advanced head")

    # copy advanced, continued head
    start_head = InterpretationHead(fun2, ctx)
    value = await start_head.continue_()
    await start_head.advance(value)
    await start_head.continue_()
    copy_start_head = start_head.copy()
    
    assert(copy_start_head.trace == [("a",), ContinueOnly])
    assert(copy_start_head._internal_trace == [])
    await assert_remaining_trace(copy_start_head, ["b", "c"], "copied advanced head")

    # double copy advanced, continued head
    start_head = InterpretationHead(fun2, ctx)
    value = await start_head.continue_()
    await start_head.advance(value)
    await start_head.continue_()
    copy_copy_start_head = start_head.copy().copy()
    
    assert(copy_copy_start_head.trace == [("a",), ContinueOnly])
    assert(copy_copy_start_head._internal_trace == [])
    await assert_remaining_trace(copy_copy_start_head, ["b", "c"], "copied advanced head")

    # copy advanced, continued head but advance prototype first
    start_head = InterpretationHead(fun2, ctx)
    value = await start_head.continue_()
    await start_head.advance(value)
    value = await start_head.continue_()

    copy_start_head = start_head.copy()
    await start_head.advance(value)

    assert(copy_start_head.trace == [("a",), ContinueOnly])
    assert(copy_start_head._internal_trace == [])
    await assert_remaining_trace(copy_start_head, ["b", "c"], "copied advanced head")
    
    assert start_head._internal_trace == [("a",), ("b",)]

if __name__ == "__main__":
    asyncio.run(test_multi_head())


import inspect
from dataclasses import dataclass
from typing import Any, Dict, Optional

from lmql.ops.ops import *
from lmql.runtime.output_writer import silent
from lmql.runtime.prompt_interpreter import PromptInterpreter, LMQLResult
from lmql.runtime.model_registry import LMQLModelRegistry
from lmql.runtime.postprocessing.conditional_prob import ConditionalDistributionPostprocessor
from lmql.runtime.postprocessing.group_by import GroupByPostprocessor
from lmql.runtime.langchain import LMQLChainMixIn

def register_model(identifier, ModelClass):
    LMQLModelRegistry.registry[identifier] = ModelClass

class LMQLInputVariableScope:
    def __init__(self, f, calling_frame):
        self.fct = f
        
        self.builtins = __builtins__
        self.globals =  calling_frame.frame.f_globals
        self.locals = calling_frame.frame.f_locals
    
    def resolve(self, name):
        if name in self.locals.keys():
            return self.locals[name]
        elif name in self.globals.keys():
            return self.globals[name]
        elif name in self.builtins.keys():
            return self.builtins[name]
        else:
            return None
            # assert False, "Failed to resolve variable '" + name + "' in @lmql.query " + str(self.fct)

@dataclass
class FunctionContext:
    argnames: List[str]
    args_of_query: List[str]
    scope: LMQLInputVariableScope


class LMQLQueryFunction(LMQLChainMixIn):
    fct: Any
    output_variables: List[str]
    postprocessors: List[Any]
    scope: Any

    output_writer: Optional[Any] = None
    args: Optional[List[str]] = None
    model: Optional[Any] = None
    function_context: Optional[FunctionContext] = None

    is_langchain_use: bool = False
    
    def __init__(self, fct, output_variables, postprocessors, scope, *args, **kwargs):
        # check for pydantic base class and do kw initialization then
        if hasattr(self, "schema_json"):
            super().__init__(fct=fct, output_variables=output_variables, postprocessors=postprocessors, scope=scope, *args, **kwargs)
        else:
            # otherwise, do standard initialization
            self.fct = fct
            self.output_variables = output_variables
            self.postprocessors = postprocessors
            self.scope = scope
        
        self.output_writer = None
        self.args = [a for a in inspect.getfullargspec(fct).args if a != "context"]
        self.model = None
        # only set if the query is defined inline of a Python file
        self.function_context = None

    @property
    def input_keys(self) -> List[str]:
        self.is_langchain_use = True
        return self.args
    
    def __getattribute__(self, __name: str) -> Any:
        return super().__getattribute__(__name)

    @property
    def output_keys(self) -> List[str]:
        return self.output_variables

    def force_model(self, model):
        self.model = model

    def make_kwargs(self, *args, **kwargs):
        if self.function_context is None:
            return kwargs
        else:
            argnames = self.function_context.argnames
            args_of_query = self.function_context.args_of_query
            scope = self.function_context.scope

        assert len(args) == len(argnames), f"@lmql.query {self.fct.__name__} expects {len(argnames)} positional arguments, but got {len(args)}."
        captured_variables = set(args_of_query)
        for name, value in zip(argnames, args):
            if name in args_of_query:
                kwargs[name] = value
                captured_variables.remove(name)
    
        for v in captured_variables:
            kwargs[v] = scope.resolve(v)
        
        if "output_writer" in kwargs:
            self.output_writer = kwargs["output_writer"]
            del kwargs["output_writer"]
        else:
            self.output_writer = silent

        return kwargs

    def __call__(self, *args, **kwargs):
        if not self.is_langchain_use:
            return self.__acall__(*args, **kwargs)
        else:
            return super().__call__(*args, **kwargs)

    async def __acall__(self, *args, **kwargs):
        kwargs = self.make_kwargs(*args, **kwargs)

        interpreter = PromptInterpreter(force_model=self.model)
        if self.output_writer is not None:
            interpreter.output_writer = self.output_writer

        query_kwargs = {}
        for a in self.args:
            if a in kwargs.keys():
                query_kwargs[a] = kwargs[a]
            else:
                query_kwargs[a] = self.scope.resolve(a)
        
        # execute main prompt
        results = await interpreter.run(self.fct, **query_kwargs)

        # applies distribution postprocessor if required
        results = await (ConditionalDistributionPostprocessor(interpreter).process(results))

        # apply remaining postprocessors
        if self.postprocessors is not None:
            for postprocessor in self.postprocessors:
                results = await postprocessor.process(results, self.output_writer)
        
        interpreter.print_stats()

        return results

def tag(t):
    return f"<lmql:{t}/>"

def compiled_query(output_variables=None, group_by=None):
    if output_variables is None:
        output_variables = []
    
    postprocessors = []
    
    calling_frame = inspect.stack()[1]
    
    if group_by is not None:
        postprocessors.append(GroupByPostprocessor(group_by))
    
    # TODO validate that only one postprocessor is used

    def func_transformer(fct):
        return LMQLQueryFunction(fct, 
                                 output_variables=output_variables, 
                                 postprocessors=postprocessors, 
                                 scope=LMQLInputVariableScope(fct, calling_frame))
    return func_transformer


import sys
import json
import asyncio
import inspect

def send(*args):
    print(*args)
    sys.stdout.flush()

class LiveApp:
    def __init__(self, name, fct, client_script=None, client_html=None):
        self.name = name
        self.fct = fct

        if client_script is not None:
            with open(client_script, "r") as f:
                self.client_script = f.read()
        else:
            self.client_script = None

        if client_html is not None:
            with open(client_html, "r") as f:
                self.html = f.read()
        else:
            self.html = None

    def __repr__(self) -> str:
        return "<LiveApp: {}>".format(self.name)

    @staticmethod
    def kill():
        LiveApp.interrupt = True


    @staticmethod
    async def ainput(string: str = "", web=False) -> str:
        # on the web, we do not have stdin and use a custom input mechanism
        if web:
            fut = asyncio.get_event_loop().create_future()
            LiveApp.stdin_queue.put_nowait(fut)
            return await fut
        return (await asyncio.get_event_loop().run_in_executor(
                None, sys.stdin.readline)).rstrip("\n")

    @staticmethod
    async def send_input(data):
        try:
            fut = LiveApp.stdin_queue.get_nowait()
            fut.set_result(data)
        except asyncio.QueueEmpty:
            raise Exception("No input request pending")
    
    @staticmethod
    def cli():
        return asyncio.run(LiveApp.async_cli())

    @staticmethod
    async def async_cli(args=None):
        if args is None:
            args = sys.argv

        if "endpoints" in set(args):
            for name,app in LiveApp.endpoint.items():
                print(app.name)
        elif "client-script" in set(args):
            app_name = args[2]
            for name,app in LiveApp.endpoint.items():
                if app.name == app_name:
                    if app.client_script is not None: 
                        print(app.client_script)
                    else: 
                        print("")
        elif "client-html" in set(args):
            app_name = args[2]
            for name,app in LiveApp.endpoint.items():
                if app.name == app_name:
                    if app.html is not None: 
                        print(app.html)
                    else: 
                        print("")
        else:
            endpoint = LiveApp.endpoint[args[1]]
            input = json.loads(args[2])
            endpoint_args = json.loads(args[3])
            
            # check if fct has kwarg web
            if args[0] == "web" and "web" in inspect.signature(endpoint.fct).parameters:
                kwargs = {"web": True}
            else:
                kwargs = {}
            
            result = await endpoint.fct(input, *endpoint_args, **kwargs)
            if result is not None: print(result)

LiveApp.endpoint = {}
LiveApp.stdin_queue = asyncio.Queue()
LiveApp.interrupt = False

class Int64JSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, int):
            return int(obj)
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            return str(obj)

def add_debugger_output(type_identifier: str, json_encodable_dict: dict):
    data = {
        "type": type_identifier,
        "data": json_encodable_dict
    }
    payload = json.dumps(data, cls=Int64JSONEncoder)
    
    print("DEBUGGER OUTPUT", payload.replace("\n", "\\n"), flush=True)

# live annotation registers a function with LiveApp.endpoint
# annotation arguments are also functions 'input' and 'output'
def live(client_script=None, client_html=None):
    def func_transformer(f):
        LiveApp.endpoint[f.__name__] = LiveApp(f.__name__, f, client_script=client_script, client_html=client_html)
        return f
    return func_transformer


# from lmql.tests.expr_test_utils import *
from lmql.ops.ops import Sentences, NextToken
from lmql.ops.token_set import VocabularyMatcher
from transformers import AutoTokenizer

# enable_show_transformed()

def test_sentences_op():
    VocabularyMatcher.init(AutoTokenizer.from_pretrained("gpt2"))

    op = Sentences([])

    assert op.forward("This is a test. Another sentence. This is incomplete") == ('This is a test.', ' Another sentence.', ' This is incomplete')

    assert str(op.follow("This is a test" + NextToken)) == "{eos} -> ('This is a test',) and * -> ('This is a test<lmql.next>',)"
    assert str(op.follow("This is a test." + NextToken)) == "{eos} -> ('This is a test.',) and * -> ('This is a test.', '<lmql.next>')"



test_sentences_op()


# coding=utf-8
# This file is adapted from the original file as part of the huggingface/transformers library, 
# which is licensed and distributed according to the following terms:
#
# Copyright 2020 The HuggingFace Inc. team
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from abc import ABC, abstractmethod
from collections import UserDict
from typing import Optional, Tuple

import torch

from transformers.file_utils import add_start_docstrings


PROCESS_INPUTS_DOCSTRING = r"""
    Args:
        input_ids (:obj:`torch.LongTensor` of shape :obj:`(batch_size * num_beams, sequence_length)`):
            Indices of input sequence tokens in the vocabulary.

            Indices can be obtained using any class inheriting from :class:`~transformers.PreTrainedTokenizer`. See
            :meth:`transformers.PreTrainedTokenizer.encode` and :meth:`transformers.PreTrainedTokenizer.__call__` for
            details.

            `What are input IDs? <../glossary.html#input-ids>`__
        next_scores (:obj:`torch.FloatTensor` of shape :obj:`(batch_size, 2 * num_beams)`):
            Current scores of the top :obj:`2 * num_beams` non-finished beam hypotheses.
        next_tokens (:obj:`torch.LongTensor` of shape :obj:`(batch_size, 2 * num_beams)`):
            :obj:`input_ids` of the tokens corresponding to the top :obj:`2 * num_beams` non-finished beam hypotheses.
        next_indices (:obj:`torch.LongTensor` of shape :obj:`(batch_size, 2 * num_beams)`):
            Beam indices indicating to which beam hypothesis the :obj:`next_tokens` correspond.
        pad_token_id (:obj:`int`, `optional`):
            The id of the `padding` token.
        eos_token_id (:obj:`int`, `optional`):
            The id of the `end-of-sequence` token.

    Return:
        :obj:`UserDict`: A dictionary composed of the fields as defined above:

            - **next_beam_scores** (:obj:`torch.FloatTensor` of shape :obj:`(batch_size * num_beams)`) -- Updated
              scores of all non-finished beams.
            - **next_beam_tokens** (:obj:`torch.FloatTensor` of shape :obj:`(batch_size * num_beams)`) -- Next tokens
              to be added to the non-finished beam_hypotheses.
            - **next_beam_indices** (:obj:`torch.FloatTensor` of shape :obj:`(batch_size * num_beams)`) -- Beam indices
              indicating to which beam the next tokens shall be added.

"""

FINALIZE_INPUTS_DOCSTRING = r"""
    Args:
        input_ids (:obj:`torch.LongTensor` of shape :obj:`(batch_size * num_beams, sequence_length)`):
            Indices of input sequence tokens in the vocabulary.

            Indices can be obtained using any class inheriting from :class:`~transformers.PreTrainedTokenizer`. See
            :meth:`transformers.PreTrainedTokenizer.encode` and :meth:`transformers.PreTrainedTokenizer.__call__` for
            details.

            `What are input IDs? <../glossary.html#input-ids>`__
        final_beam_scores (:obj:`torch.FloatTensor` of shape :obj:`(batch_size * num_beams)`):
            The final scores of all non-finished beams.
        final_beam_tokens (:obj:`torch.FloatTensor` of shape :obj:`(batch_size * num_beams)`):
            The last tokens to be added to the non-finished beam_hypotheses.
        final_beam_indices (:obj:`torch.FloatTensor` of shape :obj:`(batch_size * num_beams)`):
            The beam indices indicating to which beam the :obj:`final_beam_tokens` shall be added.
        pad_token_id (:obj:`int`, `optional`):
            The id of the `padding` token.
        eos_token_id (:obj:`int`, `optional`):
            The id of the `end-of-sequence` token.

    Return:
        :obj:`torch.LongTensor` of shape :obj:`(batch_size * num_return_sequences, sequence_length)`: The generated
        sequences. The second dimension (sequence_length) is either equal to :obj:`max_length` or shorter if all
        batches finished early due to the :obj:`eos_token_id`.

"""


class BeamScorer(ABC):
    """
    Abstract base class for all beam scorers that are used for :meth:`~transformers.PreTrainedModel.beam_search` and
    :meth:`~transformers.PreTrainedModel.beam_sample`.
    """

    @abstractmethod
    @add_start_docstrings(PROCESS_INPUTS_DOCSTRING)
    def process(
        self,
        input_ids: torch.LongTensor,
        next_scores: torch.FloatTensor,
        next_tokens: torch.LongTensor,
        next_indices: torch.LongTensor,
        **kwargs
    ) -> Tuple[torch.Tensor]:
        raise NotImplementedError("This is an abstract method.")

    @abstractmethod
    @add_start_docstrings(FINALIZE_INPUTS_DOCSTRING)
    def finalize(
        self,
        input_ids: torch.LongTensor,
        next_scores: torch.FloatTensor,
        next_tokens: torch.LongTensor,
        next_indices: torch.LongTensor,
        max_length: int,
        **kwargs
    ) -> torch.LongTensor:
        raise NotImplementedError("This is an abstract method.")


class BeamSearchScorer(BeamScorer):
    r"""
    :class:`transformers.BeamScorer` implementing standard beam search decoding.

    Adapted in part from `Facebook's XLM beam search code
    <https://github.com/facebookresearch/XLM/blob/9e6f6814d17be4fe5b15f2e6c43eb2b2d76daeb4/src/model/transformer.py#L529>`__.

    Reference for the diverse beam search algorithm and implementation `Ashwin Kalyan's DBS implementation
    <https://github.com/ashwinkalyan/dbs/blob/master/dbs/beam_utils.lua>`__

    Args:
        batch_size (:obj:`int`):
            Batch Size of :obj:`input_ids` for which standard beam search decoding is run in parallel.
        max_length (:obj:`int`):
            The maximum length of the sequence to be generated.
        num_beams (:obj:`int`):
            Number of beams for beam search.
        device (:obj:`torch.device`):
            Defines the device type (*e.g.*, :obj:`"cpu"` or :obj:`"cuda"`) on which this instance of
            :obj:`BeamSearchScorer` will be allocated.
        length_penalty (:obj:`float`, `optional`, defaults to 1.0):
            Exponential penalty to the length. 1.0 means no penalty. Set to values < 1.0 in order to encourage the
            model to generate shorter sequences, to a value > 1.0 in order to encourage the model to produce longer
            sequences.
        do_early_stopping (:obj:`bool`, `optional`, defaults to :obj:`False`):
            Whether to stop the beam search when at least ``num_beams`` sentences are finished per batch or not.
        num_beam_hyps_to_keep (:obj:`int`, `optional`, defaults to 1):
            The number of beam hypotheses that shall be returned upon calling
            :meth:`~transformer.BeamSearchScorer.finalize`.
        num_beam_groups (:obj:`int`):
            Number of groups to divide :obj:`num_beams` into in order to ensure diversity among different groups of
            beams. See `this paper <https://arxiv.org/pdf/1610.02424.pdf>`__ for more details.
    """

    def __init__(
        self,
        batch_size: int,
        num_beams: int,
        device: torch.device,
        length_penalty: Optional[float] = 1.0,
        do_early_stopping: Optional[bool] = False,
        num_beam_hyps_to_keep: Optional[int] = 1,
        num_beam_groups: Optional[int] = 1,
        **kwargs,
    ):
        self.num_beams = num_beams
        self.device = device
        self.length_penalty = length_penalty
        self.do_early_stopping = do_early_stopping
        self.num_beam_hyps_to_keep = num_beam_hyps_to_keep
        self.num_beam_groups = num_beam_groups
        self.group_size = self.num_beams // self.num_beam_groups

        self._is_init = False
        self._beam_hyps = [
            BeamHypotheses(
                num_beams=self.num_beams,
                length_penalty=self.length_penalty,
                early_stopping=self.do_early_stopping,
            )
            for _ in range(batch_size)
        ]
        self._done = torch.tensor([False for _ in range(batch_size)], dtype=torch.bool, device=self.device)

        if not isinstance(num_beams, int) or num_beams <= 1:
            raise ValueError(
                f"`num_beams` has to be an integer strictly greater than 1, but is {num_beams}. For `num_beams` == 1, one should make use of `greedy_search` instead."
            )

        if not isinstance(num_beam_groups, int) or (num_beam_groups > num_beams) or (num_beams % num_beam_groups != 0):
            raise ValueError(
                f"`num_beam_groups` has to be an integer smaller or equal than `num_beams` and `num_beams` "
                f"has to be divisible by `num_beam_groups`, but is {num_beam_groups} with `num_beams` being {num_beams}."
            )

        if "max_length" in kwargs:
            warnings.warn(
                "Passing `max_length` to BeamSearchScorer is deprecated and has no effect."
                "`max_length` should be passed directly to `beam_search(...)`, `beam_sample(...)`"
                ",or `group_beam_search(...)`."
            )

    @property
    def is_done(self) -> bool:
        return self._done.all()

    def process(
        self,
        input_ids: torch.LongTensor,
        next_scores: torch.FloatTensor,
        next_tokens: torch.LongTensor,
        next_indices: torch.LongTensor,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
    ) -> Tuple[torch.Tensor]:
        cur_len = input_ids.shape[-1]
        batch_size = len(self._beam_hyps)
        assert batch_size == (input_ids.shape[0] // self.group_size)

        device = input_ids.device
        next_beam_scores = torch.zeros((batch_size, self.group_size), dtype=next_scores.dtype, device=device)
        next_beam_tokens = torch.zeros((batch_size, self.group_size), dtype=next_tokens.dtype, device=device)
        next_beam_indices = torch.zeros((batch_size, self.group_size), dtype=next_indices.dtype, device=device)

        for batch_idx, beam_hyp in enumerate(self._beam_hyps):
            if self._done[batch_idx]:
                assert (
                    len(beam_hyp) >= self.num_beams
                ), f"Batch can only be done if at least {self.num_beams} beams have been generated"
                assert (
                    eos_token_id is not None and pad_token_id is not None
                ), "generated beams >= num_beams -> eos_token_id and pad_token have to be defined"
                # pad the batch
                next_beam_scores[batch_idx, :] = 0
                next_beam_tokens[batch_idx, :] = pad_token_id
                next_beam_indices[batch_idx, :] = 0
                continue

            # next tokens for this sentence
            beam_idx = 0
            for beam_token_rank, (next_token, next_score, next_index) in enumerate(
                zip(next_tokens[batch_idx], next_scores[batch_idx], next_indices[batch_idx])
            ):
                batch_beam_idx = batch_idx * self.group_size + next_index
                # add to generated hypotheses if end of sentence
                if (eos_token_id is not None) and (next_token.item() == eos_token_id):
                    # if beam_token does not belong to top num_beams tokens, it should not be added
                    is_beam_token_worse_than_top_num_beams = beam_token_rank >= self.group_size
                    if is_beam_token_worse_than_top_num_beams:
                        continue
                    beam_hyp.add(
                        input_ids[batch_beam_idx].clone(),
                        next_score.item(),
                    )
                else:
                    # add next predicted token since it is not eos_token
                    next_beam_scores[batch_idx, beam_idx] = next_score
                    next_beam_tokens[batch_idx, beam_idx] = next_token
                    next_beam_indices[batch_idx, beam_idx] = batch_beam_idx
                    beam_idx += 1

                # once the beam for next step is full, don't add more tokens to it.
                if beam_idx == self.group_size:
                    break

            if beam_idx < self.group_size:
                raise ValueError(
                    f"At most {self.group_size} tokens in {next_tokens[batch_idx]} can be equal to `eos_token_id: {eos_token_id}`. Make sure {next_tokens[batch_idx]} are corrected."
                )

            # Check if we are done so that we can save a pad step if all(done)
            self._done[batch_idx] = self._done[batch_idx] or beam_hyp.is_done(
                next_scores[batch_idx].max().item(), cur_len
            )

        return UserDict(
            {
                "next_beam_scores": next_beam_scores.view(-1),
                "next_beam_tokens": next_beam_tokens.view(-1),
                "next_beam_indices": next_beam_indices.view(-1),
            }
        )

    def finalize(
        self,
        input_ids: torch.LongTensor,
        final_beam_scores: torch.FloatTensor,
        final_beam_tokens: torch.LongTensor,
        final_beam_indices: torch.LongTensor,
        max_length: int,
        pad_token_id: Optional[int] = None,
        eos_token_id: Optional[int] = None,
    ) -> Tuple[torch.LongTensor]:
        batch_size = len(self._beam_hyps)

        # finalize all open beam hypotheses and add to generated hypotheses
        for batch_idx, beam_hyp in enumerate(self._beam_hyps):
            if self._done[batch_idx]:
                continue

            # all open beam hypotheses are added to the beam hypothesis
            # beam hypothesis class automatically keeps the best beams
            for beam_id in range(self.num_beams):
                batch_beam_idx = batch_idx * self.num_beams + beam_id
                final_score = final_beam_scores[batch_beam_idx].item()
                final_tokens = input_ids[batch_beam_idx]
                beam_hyp.add(final_tokens, final_score)

        # select the best hypotheses
        sent_lengths = input_ids.new(batch_size * self.num_beam_hyps_to_keep)
        best = []
        best_scores = torch.zeros(batch_size * self.num_beam_hyps_to_keep, device=self.device, dtype=torch.float32)

        # retrieve best hypotheses
        for i, beam_hyp in enumerate(self._beam_hyps):
            sorted_hyps = sorted(beam_hyp.beams, key=lambda x: x[0])
            for j in range(self.num_beam_hyps_to_keep):
                best_hyp_tuple = sorted_hyps.pop()
                best_score = best_hyp_tuple[0]
                best_hyp = best_hyp_tuple[1]
                sent_lengths[self.num_beam_hyps_to_keep * i + j] = len(best_hyp)

                # append to lists
                best.append(best_hyp)
                best_scores[i * self.num_beam_hyps_to_keep + j] = best_score

        # prepare for adding eos
        sent_max_len = min(sent_lengths.max().item() + 1, max_length)
        decoded: torch.LongTensor = input_ids.new(batch_size * self.num_beam_hyps_to_keep, sent_max_len)
        # shorter batches are padded if needed
        if sent_lengths.min().item() != sent_lengths.max().item():
            assert pad_token_id is not None, "`pad_token_id` has to be defined"
            decoded.fill_(pad_token_id)

        # fill with hypotheses and eos_token_id if the latter fits in
        for i, hypo in enumerate(best):
            decoded[i, : sent_lengths[i]] = hypo
            if sent_lengths[i] < max_length:
                decoded[i, sent_lengths[i]] = eos_token_id
        return UserDict(
            {
                "sequences": decoded,
                "sequence_scores": best_scores,
            }
        )


class BeamHypotheses:
    def __init__(self, num_beams: int, length_penalty: float, early_stopping: bool):
        """
        Initialize n-best list of hypotheses.
        """
        self.length_penalty = length_penalty
        self.early_stopping = early_stopping
        self.num_beams = num_beams
        self.beams = []
        self.worst_score = 1e9

    def __len__(self):
        """
        Number of hypotheses in the list.
        """
        return len(self.beams)

    def add(self, hyp: torch.LongTensor, sum_logprobs: float):
        """
        Add a new hypothesis to the list.
        """
        score = sum_logprobs / (hyp.shape[-1] ** self.length_penalty)
        if len(self) < self.num_beams or score > self.worst_score:
            self.beams.append((score, hyp))
            if len(self) > self.num_beams:
                sorted_next_scores = sorted([(s, idx) for idx, (s, _) in enumerate(self.beams)])
                del self.beams[sorted_next_scores[0][1]]
                self.worst_score = sorted_next_scores[1][0]
            else:
                self.worst_score = min(score, self.worst_score)

    def is_done(self, best_sum_logprobs: float, cur_len: int) -> bool:
        """
        If there are enough hypotheses and that none of the hypotheses being generated can become better than the worst
        one in the heap, then we are done with this sentence.
        """

        if len(self) < self.num_beams:
            return False
        elif self.early_stopping:
            return True
        else:
            cur_score = best_sum_logprobs / cur_len ** self.length_penalty
            ret = self.worst_score >= cur_score
            return ret



import re
from dataclasses import dataclass

@dataclass
class TemplateVariable:
    name: str

@dataclass
class DistributionVariable:
    name: str

def qstring_to_stmts(qstring):
    stmts = []
    offset = 0
    for match in re.finditer("\[[A-Za-z0-9:_]+\]", qstring):
        new_offset = match.span()[0]

        # append string preceding template var
        previous_string = qstring[offset:new_offset]
        if len(previous_string) != 0: stmts.append(previous_string)

        name = qstring[match.span()[0]: match.span()[1]][1:-1]
        if name.startswith("distribution:"):
            stmts.append(DistributionVariable(name[len("distribution:"):]))
        else:
            # append template var
            stmts.append(TemplateVariable(name))
        
        # advance pointer
        offset = match.span()[1]
    
    # append string suffix
    previous_string = qstring[offset:]
    if len(previous_string) != 0: stmts.append(previous_string)

    return stmts


from lmql.runtime.decoder_head import DecoderHead
from lmql.runtime.rewriter import InputIdRewriter
import lmql.runtime.bopenai as openai
from dataclasses import dataclass
from typing import List, Any
from lmql.runtime.stats import Stats
import inspect

import lmql.runtime.dclib.decoders as decoders
import lmql.runtime.dclib as dc
from lmql.runtime.interrupt import interrupt

import asyncio

import numpy as np
from lmql.utils import nputil

class _DCLibDebugPrinter: pass
_DCLibDebugPrinter.printer = None

def set_dclib_debug_printer(printer):
    _DCLibDebugPrinter.printer = printer

# obsolete
class DcLibMaskLogitsProcessor:
    def __init__(self, n, tokenize, detokenize, eos_token_id, mask_logits_processor):
        self.n = n
        self.tokenize = tokenize
        self.detokenize = detokenize
        self.eos_token_id = eos_token_id
        self.mask_logits_processor = mask_logits_processor

    async def __call__(self, prompt_ids, next_token_logits, additional_logits_processor_mask=None):
        async def apply_processor(i, head_input_ids, head_next_token_logits):
            if additional_logits_processor_mask is not None and not additional_logits_processor_mask[i]:
                return head_next_token_logits
            
            return await self.mask_logits_processor(DecoderHead(i, 
                head_input_ids, -1, head_next_token_logits, 
                self.tokenize,
                self.detokenize,
                self.eos_token_id,
                self.n))
        
        tasks = [apply_processor(i, head_input_ids, head_next_token_logits) for i, (head_input_ids, head_next_token_logits) in enumerate(zip(prompt_ids, next_token_logits))]

        return np.stack(await asyncio.gather(*tasks), axis=0)

@dataclass
class DcLibTokenMaskResult:
    logits_mask: np.ndarray
    user_data: List[Any]

class DcLibMaskLogitsProcessorWithUserData:
    def __init__(self, vocab_size, n, tokenize, detokenize, eos_token_id, mask_logits_processor):
        self.vocab_size = vocab_size
        self.n = n
        self.tokenize = tokenize
        self.detokenize = detokenize
        self.eos_token_id = eos_token_id
        self.mask_logits_processor = mask_logits_processor

    async def __call__(self, input_ids, user_data=None, additional_logits_processor_mask=None):
        async def apply_processor(i, head_input_ids, head_next_token_logits) -> DcLibTokenMaskResult:
            sequence_user_data = user_data[i] if user_data is not None else None

            if additional_logits_processor_mask is not None and not additional_logits_processor_mask[i]:
                return DcLibTokenMaskResult(head_next_token_logits, sequence_user_data)

            logits_mask, updated_user_data = await self.mask_logits_processor(DecoderHead(i, 
                head_input_ids, -1, head_next_token_logits, 
                self.tokenize,
                self.detokenize,
                self.eos_token_id,
                self.n,
                user_data=sequence_user_data), return_user_data=True)

            return DcLibTokenMaskResult(logits_mask, updated_user_data)
        
        tasks = [apply_processor(i, ids, np.zeros(self.vocab_size, dtype=np.float32)) for i, ids in enumerate(input_ids)]
        tasks = await asyncio.gather(*tasks)

        return DcLibTokenMaskResult(np.stack([t.logits_mask for t in tasks], axis=0), [t.user_data for t in tasks])

class QueryDcLibAdapter:
    def __init__(self, vocab_size, tokenize, detokenize, bos_token_id, eos_token_id):
        self.vocab_size = vocab_size
        self.tokenize = tokenize
        self.detokenize = detokenize
        
        self.bos_token_id = bos_token_id
        self.eos_token_id = eos_token_id

        dc.set_dclib_tokenizer(dc.tokenizer("lmql-adapter-tokenizer", self.tokenize, self.detokenize, self.bos_token_id, self.eos_token_id))

    def make_decoder_head(self, i, initial_prompt_offset, s: dc.DecoderSequence):
        return DecoderHead(i, 
            s.input_ids, 0, 0,
            self.tokenize,
            self.detokenize,
            self.eos_token_id,
            initial_prompt_offset,
            user_data = s.user_data
        )

    async def score_distribution_values(self, prompt, values, dcmodel: dc.DcModel, **decoder_args):
        prompt_seq = dc.seq([self.bos_token_id] + await self.tokenize(prompt))
        value_ids = [await self.tokenize(value) for value in values]

        dcmodel.log_billable_tokens(sum(len(ids) + 1 for ids in value_ids) + len(value_ids) * (len(prompt_seq.input_ids)))
        dcmodel.log_queries(sum(len(ids) + 1 for ids in value_ids))

        value_scores = []
        for s, value in zip(await dcmodel.score([prompt_seq] * len(value_ids), value_ids), value_ids):
            s = s.expand()
            value_score = s.logprobs[-len(value):]
            value_scores.append(value_score)
        
        return value_scores

    async def query(self, prompt, mask_logits_processor, head_input_id_rewriter, active_prompt_rewriter, dcmodel, decoder_args, **kwargs):
        assert issubclass(type(dcmodel), dc.DcModel), "The provided dcmodel must be a subclass of DcModel"

        if "no_repeat_ngram_size" in decoder_args:
            print("warning: no_repeat_ngram_size is known to cause issues when used with constrained decoding, including non-termination.")

        prompt_ids = await self.tokenize(prompt)
        if self.bos_token_id is not None:
            prompt_ids = [self.bos_token_id] + prompt_ids
        n = len(prompt_ids)

        decoder_args = decoder_args.copy()

        # pass processor as decoder argument
        # decoder_args["additional_logits_processors"] = [DcLibMaskLogitsProcessor(n, self.tokenize, self.detokenize, self.eos_token_id, mask_logits_processor)]
        decoder_args["dclib_additional_logits_processor"] = DcLibMaskLogitsProcessorWithUserData(self.vocab_size, n, self.tokenize, self.detokenize, self.eos_token_id, mask_logits_processor)

        # setup prompt executing input id rewriter
        rewriter = InputIdRewriter(head_input_id_rewriter, self.tokenize, self.detokenize, self.eos_token_id, n)
        
        # pass rewriter as decoder argument
        decoder_args["input_id_rewriter"] = rewriter

        if "output_writer" in decoder_args:
            set_dclib_debug_printer(decoder_args["output_writer"])

        if _DCLibDebugPrinter.printer is not None:
            if hasattr(_DCLibDebugPrinter.printer, "records_graph"):
                if _DCLibDebugPrinter.printer.records_graph:
                    dc.set_record_graph()

        mode = decoder_args["decoder"].lower()
        decoder_fct = dc.get_decoder(mode)
        self.validate_args(decoder_args, decoder_fct)

        # alias max_length -> max_len
        if "max_length" in decoder_args:
            decoder_args["max_len"] = decoder_args["max_length"]

        # setup dcmodel for use
        dcmodel.model_args = decoder_args
        decoder_args["dcmodel"] = dcmodel
        dc.set_truncation_threshold(dcmodel.truncation_threshold)

        step_budget = decoder_args.get("step_budget", 1024)
        
        async def debug_out(decoder_step):
            if _DCLibDebugPrinter.printer is not None and dc.DecoderSequence.graph is not None:
                data = await dc.DecoderSequence.graph.json(diff=True)
                data = replace_inf_nan_with_str(data)
                _DCLibDebugPrinter.printer.add_decoder_state(data)
            dcmodel.report_stats(_DCLibDebugPrinter.printer, decoder_step)

        try:
            import time

            decoder_step = 0
            average_step_time = None
            start = time.time()
            async for _ in decoder_fct(prompt_ids, **decoder_args):
                await debug_out(decoder_step)
                decoder_step += 1

                if step_budget is not None and decoder_step >= step_budget:
                    print("warning: step budget exceeded")
                    break

                if interrupt.check():
                    interrupt.clear()
                    raise InterruptedError("lmql.runtime.interrupt")
                
                average_step_time = (time.time() - start) if average_step_time is None else (average_step_time * 0.9 + (time.time() - start) * 0.1)

                if "performance_stats" in decoder_args:
                    if decoder_step % 10 == 0:
                        Stats.print_all()
                        print("step", decoder_step, "time", average_step_time)

                start = time.time()
                
        except dc.FinishException as fe:
            # one last call to debug_out to get the final state
            await debug_out(decoder_step)
            # if dc.finish is used, the decoder sets the sequences it considers 
            # finished (return them to prompt interpreter)
            result_sequences = fe.result_sequences
            
            billable_tokens = 0
            for s in result_sequences:
                upper = len(s.input_ids)
                has_deterministic_tail = False
                while s.deterministic[upper-1]:
                    upper -= 1
                    has_deterministic_tail = True
                # +1 for the eos token
                billable_tokens += upper + (1 if has_deterministic_tail else 0)
            
            dcmodel.log_billable_tokens(billable_tokens)
            
            return [self.make_decoder_head(i,n,s) for i,s in enumerate(result_sequences)]
    
    def validate_args(self, decoder_args, decoder_fct):
        INTERNAL_ARGS = ["decoder", "dcmodel", "dclib_additional_logits_processor", "input_id_rewriter", "output_writer", "chatty_openai", "distribution_batch_size", "openai_chunksize", "step_budget", "stats", "performance_stats"]

        # get all arg names and kwarg names of decoder function
        decoder_arg_names = inspect.getfullargspec(decoder_fct).args
        decoder_kwarg_names = inspect.getfullargspec(decoder_fct).kwonlyargs
        for k in decoder_args.keys():
            if k not in decoder_arg_names and k not in decoder_kwarg_names and k not in INTERNAL_ARGS:
                raise ValueError("Unknown decoder argument: {}".format(k))

def replace_inf_nan_with_str(d):
    import math

    if type(d) is dict:
        for k, v in d.items():
            d[k] = replace_inf_nan_with_str(v)
        return d
    elif type(d) is list:
        for i, v in enumerate(d):
            d[i] = replace_inf_nan_with_str(v)
        return d
    elif type(d) is float:
        if math.isinf(d) or math.isnan(d):
            return str(d)
    return d



from tokenize import Token
from typing import Iterable, Tuple, List
from itertools import product

from lmql.ops.token_set import *
from lmql.ops.follow_map import *

lmql_operation_registry = {}

# @LMQLOp('function_name') decorator
def LMQLOp(name):
    def class_transformer(cls):
        if type(name) is list:
            for n in name:
                lmql_operation_registry[n] = f"{cls.__name__}"
            return cls
        lmql_operation_registry[name] = f"{cls.__name__}"
        return cls
    return class_transformer

class Node:
    def __init__(self, predecessors):
        assert type(predecessors) is list, "Predecessors must be a list, not {}".format(type(predecessors))
        self.predecessors = predecessors
        self.depends_on_context = False
        
        self.follow_map = None
    
    def execute_predecessors(self, trace, context):
        return [execute_op(p, trace=trace, context=context) for p in self.predecessors]

    def forward(self, *args, **kwargs):
        raise NotImplementedError(type(self) + " does not implement forward()")

    def follow(self, *args, **kwargs):
        raise NotImplementedError(type(self) + " does not implement follow()")
    
    def final(self, args, **kwargs):
        if all([a == "fin" for a in args]):
            return "fin"
        return "var"

    def __nodelabel__(self):
        return str(type(self))

def DynamicTypeDispatch(name, type_map):
    def get_handler(args):
        for signature, op in type_map:
            # fallback implementation
            if signature == "*": return op
            
            # check for matching signature
            is_match = True
            if type(signature) is tuple:
                for arg, t in zip(args, signature):
                    is_match = is_match and isinstance(arg, t)
            else:
                is_match = isinstance(args, signature)
            if is_match: return op
        raise NotImplementedError("error: no matching implemntation of {} for arguments of type {}".format(name, [type(arg) for arg in args]))
    
    class TypeDispatchingNode(Node):
        def forward(self, *args, **kwargs):
            return get_handler(args).forward(self, *args, **kwargs)
        
        def follow(self, *args, **kwargs):
            return get_handler(args).follow(self, *args, **kwargs)
        
        def final(self, *args, **kwargs):
            return get_handler(args).final(self, *args, **kwargs)
        
        def __str__(self):
            return f"<{name}>"
        
        def __repr__(self):
            return f"<{name}>"
        
        def __nodelabel__(self):
            return name

    return TypeDispatchingNode

NextToken = "<lmql.next>"

def is_next_token(t): 
    return t == NextToken

def strip_next_token(x):
    if type(x) is list:
        return [i for i in x if not is_next_token(i)]
    elif type(x) is tuple:
        return tuple(i for i in x if not is_next_token(i))
    if x.endswith(NextToken):
        x = x[:-len(NextToken)]
    return x

@LMQLOp("SENTENCES")
class Sentences(Node):
    def forward(self, v):
        sentences = tuple(self.split(v, separator=["."]))
        return self.strip(sentences)
    
    def strip(self, sentences):
        if len(sentences) == 0:
            return sentences
        elif sentences[-1] == ():
            return tuple(sentences[:-1])
        else: 
            return tuple(sentences)

    def add_end(self, stc, end):
        if len(stc) == 0: return stc
        return stc + end

    def split(self, v, separator):
        result = ()
        
        current = ""
        for c in v:
            if c in separator:
                result += (current + c,)
                current = ""
            else:
                current += c
        if len(current) > 0:
            result += (current,)
        if len(result) == 0:
            return ("",)
        return result

    def follow(self, x, **kwargs):
        v = strip_next_token(x)
        has_next_token = v != x
        sentences = tuple(self.split(v, separator=["."]))

        if has_next_token: # continues with next token
            if len(sentences) > 0 and sentences[-1].endswith("."):
                return fmap(
                    ("eos", self.strip(sentences)),
                    ("*", sentences + (NextToken,))
                )
            else:
                return fmap(
                    ("eos", self.strip(sentences)),
                    ("*", tuple(sentences[:-1] + (sentences[-1] + NextToken,)))
                )
        else:
            return fmap(
                ("*", sentences)
            )

    def final(self, x, operands=None, result=None, **kwargs):
        return x[0]
    

@LMQLOp("INT")
class IntOp(Node):
    def forward(self, x):
        if x is None: return None
        if x == "": return None

        # check int contains digits only
        if x.startswith(" "):
            x = x[1:]
        if not all([c in "0123456789" for c in x]):
            return False
        else:
            return True

    def follow(self, v, **kwargs):
        if v is None: return None
        
        has_next_token = v != strip_next_token(v)
        v = strip_next_token(v)

        context = kwargs.get("context", None)
        if context.runtime.prefers_compact_mask:
            number_tokens = tset("1","2","3","4","5","6","7","8","9","Ġ2","Ġ3","Ġ4","Ġ5","Ġ0","Ġ6","Ġ7","Ġ8","Ġ9","10","12","50","19","11","20","30","15","14","16","13","25","18","17","24","80","40","22","60","23","29","27","26","28","99","33","70","45","35","64","75","21","38","44","36","32","39","34","37","48","66","55","47","49","65","68","31","67","59","77","58","69","88","46","57","43","42","78","79","90","95","41","56","54","98","76","52","53","51","86","74","89","72","73","96","71","63","62","85","61","97","84","87","94","92","83","93","91","82","81", exact=True)
            number_continuation_tokens = tset("0","1","2","3","4","5","6","7","8","9","00","01","10","12","50","19","11","20","30","15","14","16","13","25","18","17","24","80","40","22","60","23","29","27","26","28","99","33","70","45","35","64","75","21","38","44","36","32","39","34","05","37","48","66","55","47","08","49","09","65","07","02","04","03","68","31","67","59","06","77","58","69","88","46","57","43","42","78","79","90","95","41","56","54","98","76","52","53","51","86","74","89","72","73","96","71","63","62","85","61","97","84","87","94","92","83","93","91","82","81", exact=True)
        else:
            number_tokens = tset("[ 1-9][0-9]*$", regex=True)
            number_continuation_tokens = tset("[0-9]+$", regex=True)

        if not has_next_token:
            return fmap(
                ("eos", len(v.strip()) != 0),
                ("*", self.forward(v))
            )

        if len(v) == 0:
            return fmap(
                (number_tokens, True),
                ("*", False)
            )
        else:
            if len(v.strip()) == 0:
                # do not allow empty strings
                return fmap(
                    (number_continuation_tokens, True),
                    ("eos", False),
                    ("*", False)
                )

            return fmap(
                (number_continuation_tokens, True),
                ("eos", True),
                ("*", False)
            )
        
    def final(self, x, operands=None, result=None, **kwargs):
        if result == False and x[0] == "inc":
            return "fin"
        return super().final(x, operands=operands, result=result, **kwargs)

@LMQLOp("TOKENS")
class TokensOp(Node):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
        self.depends_on_context = True

    def forward(self, x, context):
        import asyncio
        if x is None: return None
        if x == "": return []
        
        tokens = context.runtime.model.sync_tokenize(x)
        return tokens

    def follow(self, v, context=None, **kwargs):
        if v is None: return None
        contains_next_token = v != strip_next_token(v)
        words = self.forward(strip_next_token(v), context)

        # if len(words) > 0 and contains_next_token:
        #     # allow continuation sub-tokens
        #     continuation_tokens = tset("[^\u0120].*", regex=True)
        #     valid_continuations = union(tset("eos"), continuation_tokens)
        #     components = [((valid_continuations, (words[:-1] + [words[-1] + NextToken])))] + components
        
        return fmap(
            ("*", (words + [NextToken]) if contains_next_token else words)
        )

    def final(self, x, context, operands=None, result=None, **kwargs):
        return x[0]

@LMQLOp("WORDS")
class WordsOp(Node):
    def forward(self, x):
        if x is None: return None
        if x == "": return []
        # split on " " or "\n"
        x = x.replace(" ", " ")
        x = x.replace("\n", " ")
        x = x.replace("\u0120", " ")
        words = x.split(" ")

        return [w for w in words if w.strip() != ""]
    
    def follow(self, v, **kwargs):
        if v is None: return None
        contains_next_token = v != strip_next_token(v)
        words = self.forward(strip_next_token(v))

        components = [("*", (words + [NextToken]) if contains_next_token else words)]
        
        if len(words) > 0 and contains_next_token:
            # allow continuation sub-tokens
            continuation_tokens = tset("[^\u0120\n ].*", regex=True)
            # valid_continuations = union(tset("eos"), continuation_tokens)
            components = [
                ((continuation_tokens, (words[:-1] + [words[-1] + NextToken]))),
                ((tset("eos"), words))
            ] + components
        
        return fmap(
            *components
        )

    def final(self, x, operands=None, result=None, **kwargs):
        return x[0]

@LMQLOp("len")
class LenOp(Node):
    def forward(self, x):
        if x is None: return None
        return len(x)
    
    def follow(self, v, **kwargs):
        if v is None: return None
        if type(v) is list or type(v) is tuple:
            return len(v)
        else:
            assert type(v) is str, "len() can only be applied to strings, lists, or tuples"
            if NextToken not in v:
                return len(v)
            v = strip_next_token(v)
            
            len_masks = []
            all = "∅"
            l = 1
            while True:
                tmask = tset(charlen=l)
                all = tmask.union(all)
                # if 'all' encompasses all possible tokens, then we have enumerated all possible lengths
                if len(all) == VocabularyMatcher.instance().vocab_size:
                    break
                if len(tmask) > 0:
                    len_masks.append((tmask, len(v) + l))
                l += 1
            
            return fmap(*len_masks)

    def final(self, x, operands=None, result=None, **kwargs):
        return x[0]

class NotOp(Node):
    def forward(self, op):
        return not op

    def follow(self, v, **kwargs):
        return not v

class Lt(Node):
    def forward(self, *args):
        if any([a is None for a in args]): return None
        return args[0] < args[1]
    
    def follow(self, *args, **kwargs):
        if any([a is None for a in args]): return None
        return args[0] < args[1]

    def final(self, ops, operands=None, result=None, **kwargs):
        final_transition_indices = {"inc": 0, "dec": 1, "fin": 2, "var": 3}
        
        op1 = final_transition_indices[ops[0]]
        op2 = final_transition_indices[ops[1]]

        transition_table = [ # a < b
            # a "inc", "dec", "fin", "var"    # b
            [   "var", "fin", "fin", "var" ], # inc
            [   "var", "var", "var", "var" ], # dec
            [   "var", "fin", "fin", "var" ], # fin
            [   "var", "var", "var", "var" ], # var
        ]

        if result: 
            r = transition_table[op2][op1]
        else: 
            r = transition_table[op1][op2]
        
        return r

def Gt(preds): return Lt(list(reversed(preds)))

class EqOp(Node):
    def __init__(self, predecessors):
        super().__init__(predecessors)

    def forward(self, *args):
        return all([a == args[0] for a in args])

    def follow(self, *args, **kwargs):
        op1 = args[0]
        op2 = args[1]
        
        if op1 is None or op2 is None:
            return None
        
        if is_next_token(op1):
            if is_next_token(op2): 
                return fmap(
                    ("*", True)
                )
            else:
                return fmap(
                    (op2, True),
                    ("*", False)
                )
        if is_next_token(op2):
            if is_next_token(op1): 
                return fmap(
                    ("*", True)
                )
            else:
                return fmap(
                    (op1, True),
                    ("*", False)
                )

        if type(op1) is str or type(op1) is str:
            op_shorter = op1 if len(strip_next_token(op1)) < len(strip_next_token(op2)) else op2
            op_longer = op1 if len(strip_next_token(op1)) > len(strip_next_token(op2)) else op2

            if strip_next_token(op_longer) == op_longer and strip_next_token(op_shorter) != op_shorter:
                return InOpStrInSet([]).follow(op_shorter, [op_longer])

        return all([a == args[0] for a in args])

    def final(self, operand_final, operands=None, result=None, **kwargs):
        if not all(type(o) is str for o in operands):
            return super().final(operand_final, operands=operands, result=result, **kwargs)

        if result: # if equal, then fin iff all operands are fin
            return super().final(operand_final, operands=operands, result=result, **kwargs)
        
        if all([o == "fin" for o in operands]):
            return "fin"

        # result is False
        
        # determine longest, fixed prefix of final value
        fixed_value = None
        for o, of in zip(operands, operand_final):
            if of == "fin":
                if fixed_value is None:
                    fixed_value = o
                else:
                    if len(fixed_value) < len(o):
                        fixed_value = o
            elif of == "inc":
                if fixed_value is None: 
                    fixed_value = o
                else:
                    if len(fixed_value) < len(o):
                        fixed_value = o
            elif of == "var":
                continue
        
        # check that each operand is a prefix of the fixed_value
        for o, of in zip(operands, operand_final):
            if of == "fin":
                if fixed_value != o: return "fin"
            elif of == "inc":
                if not fixed_value.startswith(o): return "fin"
            else: # of == "var":
                continue
        
        return super().final(operand_final, operands=operands, result=result, **kwargs)

class SelectOp(Node):
    def forward(self, *args):
        if len(args[0]) <= args[1]:
            return None
        return args[0][args[1]]

    def follow(self, *args, **kwargs):
        l = args[0]
        idx = args[1]

        if l is None and idx is None:
            return None
        else:
            if l is not None and len(l) == idx + 1:
                return fmap(
                    ("eos", None),
                    ("*", l[idx])
                )
            else:
                return None

    def final(self, ops, operands, result, **kwargs):
        l = ops[0]
        idx = ops[1]

        if idx != "fin": return "var"

        if l == "fin": 
            return "fin"
        if result is not None and (l == "fin" or l == "inc"):
            return "fin"
        else: return "var"

class Var(Node):
    def __init__(self, name):
        super().__init__([])
        self.name = name

        self.depends_on_context = True
        
        # indicates whether the downstream node requires text diff information
        self.diff_aware_read = False

    async def json(self):
        return self.name

    def forward(self, context):
        if self.diff_aware_read:
            return (context.get(self.name, None), context.get_diff(self.name, None))
        return context.get(self.name, None)
    
    def follow(self, context, **kwargs):
        value = context.get(self.name, None)
        if value is None: return None
        
        # also return the text diff if required
        if self.diff_aware_read:
            value = (value, context.get_diff(self.name, None))

        # strip_next_token but also supports tuples
        def strip_nt(v):
            if type(v) is tuple: return (strip_next_token(v[0]), v[1])
            else: return strip_next_token(v)

        return fmap(
            ("eos", PredeterminedFinal(strip_nt(value), "fin")),
            ("*", value),
        )

    def final(self, x, context, operands=None, result=None, **kwargs):
        return context.final(self.name)

    def __repr__(self) -> str:
        return f"<Var {self.name}>"

class RawValueOp(Node):
    def __init__(self, args):
        super().__init__([])
        
        value, final = args
        self.value = value
        self.final_value = final

    def forward(self):
        return self.value

    def follow(self, **kwargs):
        return fmap(
            ("*", self.value)
        )

    def final(self, args, operands=None, result=None, **kwargs):
        return self.final_value

def matching_phrases_suffixes(x, allowed_phrases, allow_full_matches=False):
    x = strip_next_token(x)

    for phrase in allowed_phrases:
        if not phrase.startswith(x):
            continue
        if len(phrase) > len(x):
            yield phrase[len(x):]
        else:
            if allow_full_matches: 
                yield ""

class InOpStrInStr(Node):
    def forward(self, *args):
        if any([a is None for a in args]): return None

        return args[0] in args[1]

    def follow(self, *args, **kwargs):
        op1 = strip_next_token(args[0])
        op1_generating = args[0] != op1
        op2 = strip_next_token(args[1])
        op2_generating = args[1] != op2

        assert not op1_generating, "InOpStrInStr: left-hand side operand must not be generating"

        # if op2 is finished, then the result is fixed
        if not op2_generating: return op1 in op2
        
        # if op1 already contained
        if op1 in op2: 
            return True

        # op1 is not contained in op2, so check for partial overlap with suffixes of op2
        overlap = 0
        for i in range(len(op1)):
            if op2[-i:] == op1[:i]:
                overlap = i
        suffix = op1[overlap:]
        suffix = suffix.replace(".", r"\.").replace("*", r"\*")

        allowed_subtokens = tset(f".*{suffix}.*", regex=True)

        return fmap(
            (allowed_subtokens, True),
            (setminus("*", allowed_subtokens), False)
        )

    def final(self, op_final, result=None, **kwargs):
        if not result:
            return super().final(op_final, result=result, **kwargs)
        if op_final[1] == "inc" and op_final[0] == "fin":
            return "fin"
        return super().final(op_final, result=result, **kwargs)

class InOpStrInSet(Node):
    def forward(self, *args):
        if any([a is None for a in args]): return None
        
        x = args[0]
        allowed_phrases = args[1]
        
        if x is None: 
            return None

        for _ in matching_phrases_suffixes(x, allowed_phrases, allow_full_matches=True):
            # any match is enough
            return True
        
        return False

    def follow(self, *args, **kwargs):
        if any([a is None for a in args]): return None

        x = args[0]
        allowed_phrases = args[1]

        suffixes = list(matching_phrases_suffixes(x, allowed_phrases, allow_full_matches=True))
        num_full_matches = len([s for s in suffixes if s == ""])
        suffixes = [s for s in suffixes if s != ""]
        
        if len(suffixes) == 0:
            if num_full_matches > 0 and not x.endswith(NextToken):
                return True
            return False
        else:
            full_remainders = [s + "$" for s in suffixes]
            if num_full_matches > 0: full_remainders.append("eos")

            return fmap(
                (tset(*full_remainders), True),
                (tset(*suffixes, prefix=True), PredeterminedFinal(False, "var")),
                ("*", PredeterminedFinal(False, "fin"))
            )

    def final(self, args, operands=None, result=None, **kwargs):
        x_final = args[0]

        if result is None:
            return "var"
        elif result == False:
            if x_final == "inc" or x_final == "fin":
                return "fin"
            return "var"
        else: # result == True
            if x_final == "fin":
                return "fin"
            return "var"

InOp = DynamicTypeDispatch("InOp", (
    ((str, str), InOpStrInStr),
    ("*", InOpStrInSet),
))

class OrOp(Node):
    def forward(self, *args):
        if any([a == True for a in args]):
            return True
        elif all([a == False for a in args]):
            return False
        else:
            return None

    def follow(self, *args, **kwargs):
        return fmap(
            ("*", self.forward(*args))
        )

    def final(self, args, operands=None, result=None, **kwargs):
        if result:
            if any(a == "fin" and v == True for a,v in zip(args, operands)):
                return "fin"
            return "var"
        else: # not result
            if any(a == "var" for a in args):
                return "var"
            return "fin"

class AndOp(Node):
    def forward(self, *args):
        if type(args[0]) is tuple and len(args) == 1:
            args = args[0]

        if any([a == False for a in args]):
            return False
        elif any([a is None for a in args]):
            return None
        else:
            return all([a for a in args])

    def follow(self, *v, **kwargs):
        return fmap(
            ("*", self.forward(*v))
        )

    def final(self, args, operands=None, result=None, **kwargs):
        if result:
            if all([a == "fin" for a in args]):
                return "fin"
            return "var"
        else: # not result
            if any([a == "fin" and v == False for a,v in zip(args, operands)]):
                return "fin"
            return "var"

def seq_starts_with(seq1, seq2):
    num_matching = sum([1 if i1 == i2 else 0 for i1,i2 in zip(seq1, seq2)])
    return num_matching == len(seq2)

def remainder(seq: str, phrase: str):
    overlap = 0
    for i in range(len(phrase)):
        if seq[-i:] == phrase[:i]:
            overlap = i

    if overlap == 0: return None
    else: return phrase[i:]

@LMQLOp("STARTS_WITH")
class StartsWithOp(Node):
    def forward(self, *args):
        if any([a is None for a in args]): return None
        
        x = args[0]
        allowed_phrases = args[1]

        for phrase in allowed_phrases:
            if x.startswith(phrase):
                return True

        return False

    def follow(self, *args, **kwargs):
        if any([a is None for a in args]): return None

        x = args[0]
        allowed_phrases = args[1]

        # if there is any full match, then the result is True
        if any(strip_next_token(x).startswith(phrase) for phrase in allowed_phrases):
            return True

        # otherwise check for partial matches with the allowed phrases
        suffixes = list(matching_phrases_suffixes(x, allowed_phrases))

        if len(suffixes) == 0:
            return False
        else:
            return fmap(
                (tset(*suffixes), True),
                (ntset(*suffixes), False)
            )

    def final(self, ops_final, result=None, operands=None, **kwargs):
        op1 = ops_final[0]
        op2 = ops_final[1]

        if op1 == "inc" and op2 == "fin":
            if len(operands[0]) == 0 and not result:
                return "var"
            return "fin"
        
        return super().final(ops_final, **kwargs)

    # def final(self, args, operands=None, result=None, pattern: TokenSet=None, **kwargs):
    #     x_final = args[0]

    #     if result is None:
    #         return "var"
    #     elif result == False:
    #         x = operands[0]
    #         allowed_phrases = operands[1]

    #         if x_final == "inc":
    #             # check whether there are suffixes which could be matched
    #             suffixes = list(matching_phrases_suffixes(x, allowed_phrases))
    #             suffixes = [s for s in suffixes if pattern is None or pattern.starts_with(s)]
    #             # print("final() with x_final inc", x, allowed_phrases)
    #             if len(suffixes) > 0: return "var"
    #             else: return "fin"
    #         elif x_final == "fin":
    #             return "fin"
    #         else:
    #             return "var"
    #     else: # result == True
    #         if x_final == "inc" or x_final == "fin":
    #             return "fin"
    #         return "var"

@LMQLOp(["STOPS_AT", "stops_at"])
class StopAtOp(Node):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tokenized_stopping_phrase_cache = {}

    def execute_predecessors(self, trace, context):
        var_op: Var = self.predecessors[0]
        assert type(var_op) is Var, "The first argument of STOPS_AT must be a direct reference to a template variable."
        assert type(self.predecessors[1]) is str, "The second argument of STOPS_AT must be a string literal."
        var_op.diff_aware_read = True
        return super().execute_predecessors(trace, context)
    
    @property
    def variable(self):
        return self.predecessors[0]

    @property
    def stopping_phrase(self):
        return self.predecessors[1]

    async def stopping_phrase_tokenized(self, tokenizer):
        if tokenizer in self._tokenized_stopping_phrase_cache:
            return self._tokenized_stopping_phrase_cache[tokenizer]
        else:
            result = (await tokenizer(self.stopping_phrase))
            self._tokenized_stopping_phrase_cache[tokenizer] = result
            return result

    def forward(self, *args):
        if any([a is None for a in args]): return None

        op1, op1_diff = args[0]
        op2 = args[1]

        if op1 is None: return
        if op1_diff is None: op1_diff = ""

        matched_phrase_index = op1.rfind(op2)
        op2_in_op1 = matched_phrase_index != -1 and matched_phrase_index + len(op2) > len(op1) - len(op1_diff)

        return not op2_in_op1 or op1.endswith(op2)

    def follow(self, *args, previous_result=None, **kwargs):
        if any([a is None for a in args]): 
            return None

        op1, op1_diff = args[0]
        if op1 is None:return None
        if op1_diff is None: op1_diff = ""

        op1 = strip_next_token(op1)
        op2 = args[1]


        matched_phrase_index = op1.rfind(op2)
        op2_in_op1 = matched_phrase_index != -1 and matched_phrase_index + len(op2) > len(op1) - len(op1_diff)

        if not op2_in_op1: return fmap(("*", True))

        ends_with_stopping_phrase = op1.endswith(op2)

        if op1 != args[0][0] and ends_with_stopping_phrase:
            # print("StopAtOp.follow()", [op1], [op2], valid)
            ends_with_stopping_phrase = False
        if len(op1) == 0:
            ends_with_stopping_phrase = True
        
        return fmap(("*", ends_with_stopping_phrase))

    def final(self, ops_final, operands, result, **kwargs):
        if result: 
            if ops_final[0] == "var":
                return "var"
            return "fin"
        else: # not result
            if ops_final[0] == "var": 
                r = "var"
            elif ops_final[0] == "dec": 
                r = "var"
            else: 
                r = "fin"
            return r

class OpaqueLambdaOp(Node):
    def forward(self, *args):
        if any([a is None for a in args]): return None
        fct, *args = args
        return fct(*args)
    
    def follow(self, *v, **kwargs):
        if any([a is None for a in v]): return None

        fct, *args = v
        return fmap(
            ("*", fct(*args))
        )

def create_mask(follow_map, valid, final):
    if follow_map is None:
        return "*"
    
    allowed_tokens = tset()
    otherwise_result = None

    for pattern, result in follow_map:
        if pattern == "*":
            otherwise_result = result

        if result is not None:
            value, final = result
        else:
            value = None
            final = "var"

        if value == True or value is None:
            allowed_tokens = union(allowed_tokens,pattern)
        elif value == False and final == ("var",):
            allowed_tokens = union(allowed_tokens, pattern)
        elif value is None and len(follow_map.components) == 1:
            allowed_tokens = "*"
        elif result == (False, ('fin',)):
            if pattern != "*":
                allowed_tokens = setminus(allowed_tokens, pattern)

    if allowed_tokens == "∅":
        return tset("eos")

    if len(allowed_tokens) == 0:
        if otherwise_result is not None:
            othw_value, othw_final = otherwise_result
        else:
            othw_value, othw_final = None, "var"
        if not othw_value and othw_final == ("fin",):
            return tset("eos")
        else:
            return "*"

    return allowed_tokens

def is_node(op):
    return issubclass(type(op), Node)

def derive_final(op, trace, context, result):
    def get_final(v):
        # for nodes, get final value from trace
        if is_node(v): return trace[v][1]
        # for constants, final value is always "fin"
        return "fin"

    predecessor_final = [get_final(p) for p in op.predecessors]

    def get_predecessor_result(v):
        if is_node(v): return trace[v][0]
        return v
    
    predecessor_values = [get_predecessor_result(p) for p in op.predecessors]

    context_arg = ()
    if op.depends_on_context: 
        context_arg += (context,)
    
    return op.final(predecessor_final, *context_arg, operands=predecessor_values, result=result)

def execute_op_stops_at_only(op: Node, result=None):
    """
    Evaluates a Node and returns the list of defined StopAtOps for the query.
    """
    if result is None: result = []

    if type(op) is StopAtOp:
        result.append(op)
    elif type(op) is AndOp:
        for p in op.predecessors:
            execute_op_stops_at_only(p, result=result)
    elif type(op) is OrOp:
        # TODO: actually STOPS_AT in OR is not really supported yet
        for p in op.predecessors:
            execute_op_stops_at_only(p, result=result)
    else:
        # other ops are no-ops from a STOPS_AT perspective (cannot contain additional STOPS_AT ops)
        # TODO: what about not
        return []
    return result

def execute_op(op: Node, trace=None, context=None, return_final=False):
    # for constant dependencies, just return their value
    if not is_node(op): 
        return op
    
    # only evaluate each operation once
    if op in trace.keys(): 
        return trace[op][0]
    
    # compute predecessor values
    inputs = op.execute_predecessors(trace, context)
    
    if op.depends_on_context: 
        inputs += (context,)
    
    result = op.forward(*inputs)
    is_final = derive_final(op, trace, context, result)
    
    if trace is not None: 
        trace[op] = (result, is_final)

    if return_final:
        return result, is_final

    return result

def digest(expr, context, follow_context, no_follow=False):
    if expr is None: return True, "fin", {}

    trace = {}
    expr_value, is_final = execute_op(expr, trace=trace, context=context, return_final=True)

    if no_follow:
        return expr_value, is_final, trace

    for op, value in trace.items():
        # determine follow map of predecessors
        if len(op.predecessors) == 0: 
            # empty argtuple translates to no follow input
            intm = all_fmap((ArgTuple(), ["fin"])) 
        else:
            # use * -> value, for constant value predecessor nodes
            def follow_map(p):
                if is_node(p): return p.follow_map
                else: return fmap(("*", (p, ("fin",))))
            intm = fmap_product(*[follow_map(p) for p in op.predecessors])
        
        # apply follow map
        op_follow_map = follow_apply(intm, op, value, context=follow_context)

        # name = op.__class__.__name__
        # print(name, value)
        # print("follow({}) = {}".format(name, op_follow_map))

        setattr(op, "follow_map", op_follow_map)
    
    return expr_value, is_final, trace



import asyncio
import js
from pyodide.ffi import to_js
import json

class EndStream: pass

class PostRequest:
    def __init__(self, url, headers, payload):
        self.url = url
        self.headers = headers
        self.payload = payload

        self.queue = asyncio.Queue()
        self.content = self

        async def data_handler(error, data):
            if error: 
                self.queue.put_nowait(EndStream)
            else: 
                self.queue.put_nowait(data.encode("utf-8"))

        self.completion_task = js.openai_completion_create(url, json.dumps(payload), data_handler)

    def close(self):
        self.completion_task.cancel()

    async def iter_any(self):
        while True:
            chunk = await self.queue.get()

            if chunk is EndStream:
                break
            else:
                yield chunk

    async def __aenter__(self):
        return self
    
    async def __aexit__(self, *args):
        pass

class ClientSession:
    def post(self, url, headers, json):
        return PostRequest(url, headers, json)

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        pass



from lmql.runtime.stats import Stats

class DcGlobal: pass
DcGlobal.tokenizer = None

stats = Stats("dclib")

def get_tokenizer():
    return DcGlobal.tokenizer

def set_dclib_tokenizer(tokenizer):
    assert DcGlobal.tokenizer is None or DcGlobal.tokenizer.name == tokenizer.name, f"Cannot set dclib tokenizer to {tokenizer.name} because it is already set to {DcGlobal.tokenizer.name} (cannot use multiple tokenizers in the same process for now)"
    DcGlobal.tokenizer = tokenizer
    
def get_tokenizer():
    return DcGlobal.tokenizer


from lmql.ops.ops import InOpStrInStr, NextToken
from lmql.tests.expr_test_utils import run_all_tests
from lmql.ops.follow_map import fmap

def test_one_of():
    in_op = InOpStrInStr([])

    result = in_op.follow("abc", "ab" + NextToken)
    assert str(result) == "{c} -> True and * -> False"

    result = in_op.follow("abc", "" + NextToken)
    assert str(result) == "{abc} -> True and * -> False"

    result = in_op.follow("ade", "def" + NextToken)
    assert str(result) == "{ade} -> True and * -> False"

run_all_tests(globals())


import os
import re
import math
from dataclasses import dataclass
from lmql.ops import *
from lmql.runtime.decoder_head import DecoderHead
from lmql.runtime.output_writer import *
from lmql.runtime.rewriter import RewrittenInputIds
from lmql.runtime.model_registry import LMQLModelRegistry
from lmql.runtime.program_state import ProgramState
from lmql.runtime.backtracker import BacktrackerHeadMixin
from lmql.runtime.runtime import LMQLRuntime
from typing import Dict, Optional, Any

import numpy as np
from lmql.utils import nputil

import lmql.ops.token_set as token_set

from lmql.language.qstrings import qstring_to_stmts, TemplateVariable, DistributionVariable
from lmql.runtime.multi_head_interpretation import InterpretationHead, InterpreterCall, InterpreterHeadPool, InterpretationHeadDone

class DecodingError(Exception): ...
class FailedToDecodeVariable(Exception): ...

def is_prefix_seq(seq1, seq2):
    num_matching = 0
    for v1, v2 in zip(seq1,seq2):
        if v1 == v2: num_matching += 1
        else: break
    return num_matching > 0

@dataclass
class LMQLResult:
    prompt: str
    variables: Dict[str, str]
    
    distribution_variable: Optional[str] = None
    distribution_values: Optional[str] = None

    @property
    def requires_distribution_postprocessing(self):
        return self.distribution_variable is not None

def copy(value):
    if value is None:
        return None
    elif type(value) is str:
        return value
    elif type(value) is list:
        return value.copy()
    else:
        assert False, "unhandled type in copy()"

class LMQLContextAPI:
    def __init__(self, program_variables, interpreter):
        self.program_variables = program_variables
        self.interpreter = interpreter

    async def json(self):
        return self.program_variables

    # LMQL runtime API

    def get_var(self, name):
        return self.program_variables.get(name)

    def query(self, qstring):
        return InterpreterCall(qstring, loc=None)

    def set_model(self, model_name):
        self.interpreter.set_model(model_name)

    def set_decoder(self, method, **kwargs):
        self.interpreter.set_decoder(method, **kwargs)

    def set_where_clause(self, where):
        self.interpreter.set_where_clause(where)

    def get_all_vars(self):
        return self.program_variables.variable_values.copy()

    def set_distribution(self, distribution_variable, values):
        self.interpreter.set_distribution(distribution_variable, values)

    def get_return_value(self):
        return LMQLResult(self.prompt, self.get_all_vars(), self.interpreter.distribution_variable, self.interpreter.distribution_values)

@dataclass
class HypothesisHeadState:
    valid: bool
    final: str
    program_state: ProgramState
    mask : Any
    stopping_phrases: List[List[int]]
    full_text : str
    where: Any
    trace: Any
    num_variables_decoded: int

class HypothesisHead(BacktrackerHeadMixin, LMQLContextAPI):
    def __init__(self, query_head, interpreter):
        super().__init__(ProgramState(runtime=interpreter), interpreter)

        self.head_index = 0
        self.query_head: InterpretationHead = query_head
        self.interpreter = interpreter
    
        self.last_seen_input_ids = None
        
        # prompt actions
        self.prompt = ""
        self.prompt_stmts = []

        # current template variable that is being decoded
        self.current_variable = None
        self.current_variable_offset = -1
        self.current_variable_scores = ()
        self.recurring_variable_counter = {}

        # track progress through the prompt
        self.num_variables_decoded = 0

        # current prompt segment that is actively fed to the model (only when active prompting is used)
        self.current_prompt = None

        # marks inactive heads which are only used for copies (will be removed from the pool in the next generation)
        self.prototype = False
        # index of the prototype head that this head was originally copied from
        self.prototype_head_index = -1

        # use self as query context (e.g. to get variables)
        query_head.context = self

        # num interpreted query calls
        self.num_queries = -1

        # masks collected during backtracking (only used by BacktrackerHeadMixin, when enabled)
        self.backtracker_masks = {}

        # keep track of last state after calling where_clause_logit_processor()
        self.last_state = None

    def copy(self) -> 'HypothesisHead':
        c = HypothesisHead(self.query_head.copy(), self.interpreter)
        c.head_index = -1
        c.program_variables = self.program_variables.copy()
        
        if self.last_seen_input_ids:
            c.last_seen_input_ids = self.last_seen_input_ids.copy()
        
        c.prompt = "" + self.prompt
        c.prompt_stmts = self.prompt_stmts.copy()

        c.current_variable = self.current_variable
        c.current_variable_scores = self.current_variable_scores
        c.current_prompt = copy(self.current_prompt)
        c.current_variable_offset = self.current_variable_offset

        c.num_variables_decoded = self.num_variables_decoded

        c.recurring_variable_counter = self.recurring_variable_counter.copy()
        c.num_queries = self.num_queries
        c.backtracker_masks = self.backtracker_masks.copy()

        c.last_state = self.last_state

        return c

    # hypothesis head API

    @property
    def num_last_seen_ids(self):
        if self.last_seen_input_ids is None:
            return 0
        else:
            return len(self.last_seen_input_ids)

    async def advance(self, active_prompting=False):
        if self.current_variable is not None:
            return
        if self.current_prompt is not None and active_prompting:
            return

        async def continue_for_more_prompt_stmts():
            if len(self.prompt_stmts) != 0: return

            if self.query_head.waiting:
                if self.query_head.done: return True
                await self.query_head.continue_()
            
            if not self.query_head.done:
                assert self.num_queries < self.query_head.num_calls, "query head {} is not advanced far enough: queries {} vs. head calls {}".format(self.query_head.id, self.num_queries, self.query_head.num_calls)

                self.num_queries = self.query_head.num_calls
                qstring = self.query_head.last_call_args[0]
                self.prompt_stmts = qstring_to_stmts(qstring)
        
        if self.query_head.waiting:
            await continue_for_more_prompt_stmts()

        while self.current_variable is None and self.current_prompt is None and not self.query_head.done:
            if len(self.prompt_stmts) == 0 and self.current_variable is None:
                if self.query_head.waiting:
                    await self.query_head.continue_()
                try:
                    await self.query_head.advance(self.prompt)
                except InterpretationHeadDone:
                    # rare case that occurs when self.query_head needs to be materialized first 
                    # and then turns out to be done already
                    break
                
                await continue_for_more_prompt_stmts()

                if self.query_head.done: break

            s = self.prompt_stmts[0]

            if type(s) is str:
                if not active_prompting:
                    self.prompt += s
                    self.prompt_stmts = self.prompt_stmts[1:]
                else:
                    self.current_prompt = s
                    self.prompt_stmts = self.prompt_stmts[1:]
                    return
            elif type(s) is TemplateVariable:
                self.current_variable = s.name
                # keep track of number of times a variable with this name has been decoded
                if self.current_variable not in self.recurring_variable_counter.keys():
                    self.recurring_variable_counter[s.name] = -1
                self.recurring_variable_counter[s.name] += 1
                
                self.prompt_stmts = self.prompt_stmts[1:]
                
                return
            elif type(s) is DistributionVariable:
                # distribution variables are skipped here, as the prompt interpreter will handle them
                # self.query_head must terminate after this part of the prompt (ensure by validation)
                distribution_var = s
                self.prompt_stmts = self.prompt_stmts[1:]

                assert len(self.prompt_stmts) == 0, "Distribution variables must be the last statement in a prompt"
            else:
                assert False, "prompt interpreter encountered unsupported prompt stmt of type {}".format(type(s))
    
    async def where_clause_logit_processor(self, head: DecoderHead):
        if self.current_variable_offset == -1:
            self.current_variable_offset = head.initial_prompt_offset - 1

        is_done = await self.advance()

        # print("where()", len(head.input_ids_without_padding), [(await head.detokenize(head.input_ids_without_padding))])

        # default results (if no where clause is run)
        valid = True
        is_final = "var"
        program_variables = self.program_variables.copy()
        trace = None
        logit_mask = None

        if is_done:
            mask = tset("eos")
            logit_mask = await head.translate_mask(mask, len(head.next_token_logits))
            stopping_phrases = []
        # follow_masking disabled
        elif self.current_variable is not None:
            where = self.interpreter.where
            
            text = (await head.text(self.current_variable_offset, strip_padding=True))
            diff_text = (await head.text(max(self.current_variable_offset, self.num_last_seen_ids - 1), strip_padding=True))

            # current context
            program_variables: ProgramState = self.program_variables.copy()
            program_variables.set(self.current_variable, text, scores=self.current_variable_scores, diff=diff_text, montonicity="inc")

            # follow context
            follow_program_variables: ProgramState = self.program_variables.copy()
            follow_program_variables.set(self.current_variable, text + str(NextToken), scores=self.current_variable_scores, diff=diff_text, montonicity="inc")

            # digest token with where expr
            valid, is_final, trace = digest(where,
                context=program_variables,
                follow_context=follow_program_variables
            )

            stopping_conditions: List[StopAtOp] = execute_op_stops_at_only(where)
            stopping_phrases = {
                "tokenized": [await s.stopping_phrase_tokenized(head.tokenizer_fn) for s in stopping_conditions if s.variable.name == self.current_variable],
                "text": [s.stopping_phrase for s in stopping_conditions if s.variable.name == self.current_variable],
            }

            # check if follow_masking is disabled
            if not self.interpreter.configuration["follow_masking"]:
                logit_mask = None
                mask = "*"
            else:
                # obtain where follow map
                follow_map = where.follow_map if where is not None else None
                # print(id(self), self.current_variable, [text], "mask", follow_map)
                mask = create_mask(follow_map, valid, is_final)
                logit_mask = await head.translate_mask(mask, len(head.next_token_logits))
        else:
            assert self.current_prompt is not None, f"error: where() is called but current variable and current prompt are None and query head is {self.query_head.done}"
            next_token = await self.get_next_prompt_token(head, pop=False)
            mask = tset("<prompt>")
            stopping_phrases = []
            logit_mask = await head.make_mask(next_token, vocab_size=len(head.next_token_logits))

        await self.debugger_output(head, valid, is_final, mask, stopping_phrases, program_variables, trace)

        # also include backtracking mask (only present if backtracking is enabled)
        logit_mask_backtracker = self.get_backtracker_mask(head)
        if logit_mask_backtracker is not None: 
            if logit_mask is not None:
                logit_mask = np.logical_and(logit_mask, np.logical_not(logit_mask_backtracker))
            else:
                logit_mask = np.logical_not(logit_mask_backtracker)

        # no mask, no logits processing
        if logit_mask is None:
            return head.next_token_logits

        # detect vocabulary mismatch
        if len(logit_mask) > len(head.next_token_logits):
            # sometimes the mask is longer than the logits (due to unused tokens)
            logit_mask = logit_mask[:len(head.next_token_logits)]
        else:
            assert len(head.next_token_logits) == len(logit_mask), "Vocabulary Mismatch: The LMQL client expects the served model to have a vocabulary dimension of {} but {} was found. Is the LMQL inference API running the correct model?". format(
                len(logit_mask), len(head.next_token_logits))

        # otherwise set disallowed logits to -inf (float min)
        head.next_token_logits[np.logical_not(logit_mask)] = np.finfo(np.float32).min

        return head.next_token_logits

    async def debugger_output(self, head: DecoderHead, valid, is_final, mask, stopping_phrases, program_variables, trace):
        where = self.interpreter.where

        # more expensive but needed when there are bugs in input_id rewriting
        full_text = await head.text(strip_padding=True, strip_eos=True)

        if self.current_prompt is not None:
            variable = f"<PROMPTING>"
            where = None
            trace = {}
            valid = True
            is_final = "var"
            program_variables = self.program_variables
        elif self.current_variable is not None:
            variable_recurrence = self.recurring_variable_counter[self.current_variable]
            variable = f"{self.current_variable}[{variable_recurrence}]"
        else:
            variable = "__done__"
            where = None
            trace = {}
            valid = True
            is_final = "fin"
            program_variables = self.program_variables
        
        self.interpreter.debugger_output(variable, head.seq_idx, full_text, where, trace, valid, is_final, mask, len(head.input_ids), program_variables)

        self.last_state = HypothesisHeadState(valid, is_final, program_variables.copy(), mask, stopping_phrases, full_text, where=where, trace=trace, num_variables_decoded=self.num_variables_decoded)

    async def json(self):
        if self.current_prompt is not None:
            variable = f"<PROMPTING>"
        elif self.current_variable is not None:
            variable_recurrence = self.recurring_variable_counter[self.current_variable]
            variable = f"{self.current_variable}[{variable_recurrence}]"
        else:
            variable = "__done__"

        if self.last_state is None:
            return {
                "head_index": self.head_index,
                "variable": variable, 
                # indicates dangeling hypothesis head without known state
                "unfinished": True
            }
        else:
            from lmql.utils.graph import CytoscapeGraphWriter

            def node_data(op):
                result = "-"
                if self.last_state.trace is not None and op in self.last_state.trace:
                    result = self.last_state.trace[op]

                follow_map = "-"
                if hasattr(op, "follow_map"):
                    follow_map = str(op.follow_map)
                return {
                    "result": result,
                    "follow_map": follow_map,
                    "repr": repr(op)
                }

            writer = CytoscapeGraphWriter(extra_data_provider=node_data)
            writer.write(self.last_state.where)
            
            last_state_dict = self.last_state.__dict__.copy()
            del last_state_dict["where"]
            del last_state_dict["trace"]

            last_state_dict["where"] = writer.graph.to_json(return_dict=True)

            return {
                "head_index": self.head_index,
                "variable": variable,
                # include information on valid, final, trace, mask, full_text
                **last_state_dict,
                "program_variables": self.program_variables
            }

    async def prepare_prompt(self, current_prompt, head: DecoderHead):
        if type(current_prompt) is list:
            return current_prompt
        else:
            assert type(current_prompt) is str
            return await head.tokenize(current_prompt)

    async def active_prompt(self, head: DecoderHead):
        assert self.interpreter.configuration["backtracking"] == False, "backtracking is not supported with active prompting (beam search)"

        # print("Head", head.head, "active_prompt()", len(head.input_ids_without_padding), [(await head.text(strip_padding=True))])
        if self.current_variable_offset == -1:
            self.current_variable_offset = head.initial_prompt_offset

        # update last seen ids
        full_ids = head.input_ids_without_padding.tolist() + [head.next_token_id]
        self.last_seen_input_ids = full_ids

        # print("active_prompt()", len(full_ids), [(await head.detokenize(full_ids))], head.next_token_id, self.current_prompt is not None)

        if head.is_at_eos() and self.current_variable is not None:
            text = (await head.text(strip_padding=True, offset=self.current_variable_offset))
            self.program_variables.set(self.current_variable, text, scores=self.current_variable_scores, diff="", montonicity="fin")
            self.current_variable = None
            self.num_variables_decoded += 1
            self.current_variable_scores = ()

            await self.advance(active_prompting=True)
            
            # if query is finished, keep eos token (indicates to decoder that this sequence/query has finished decoding)
            if self.query_head.done: return None
            # assert not head.is_at_eos(), "active_prompt() called when head is at eos but current variable is None"
            # return None

        if self.current_prompt is not None:
            next_token = await self.get_next_prompt_token(head, pop=True)
            if next_token is None: 
                return
            
            full_ids = head.input_ids_without_padding.tolist() + [next_token]
            self.last_seen_input_ids = full_ids

            self.current_variable_offset = len(head.input_ids_without_padding) + 1
            return next_token

    async def get_next_prompt_token(self, head: DecoderHead, pop=True):
        if self.current_prompt is not None:
            self.current_prompt = await self.prepare_prompt(self.current_prompt, head)
            
            # if current prompt is empty, try again with next prompt statement (var or prompt)
            if len(self.current_prompt) == 0:
                self.current_prompt = None
                return await self.get_next_prompt_token(head)
            else:
                if not pop:
                    return self.current_prompt[0]
                next_token = self.current_prompt.pop(0)

                if len(self.current_prompt) == 0:
                    self.current_prompt = None
                    await self.advance(active_prompting=True)

                return next_token
        return None

    async def head_input_id_rewriter(self, head: DecoderHead):
        """
        Rewrite input IDs of this decoder head. 
        
        Returns None, if no rewrite is required.
        """
        if self.current_variable_offset == -1:
            self.current_variable_offset = head.initial_prompt_offset

        full_ids = head.input_ids_without_padding.tolist() + [head.next_token_id]
        self.last_seen_input_ids = full_ids

        if self.current_variable is not None:
            # keep track of scores per decoded variable
            if head.next_token_logprob is not None:
                self.current_variable_scores += (head.next_token_logprob,)
            else:
                self.current_variable_scores += (nputil.log_softmax(head.next_token_logits)[head.next_token_id],)

            # backtracking if enabled
            if self.interpreter.configuration["backtracking"] == True:
                # backtrack if required
                if not await self.backtracking_check_valid(head):
                    full_ids = head.input_ids_without_padding.tolist()[:-1]
                    self.last_seen_input_ids = full_ids
                    return RewrittenInputIds(appended_input_ids=None, strip_eos=-2)

        if head.is_at_eos() and self.current_variable is not None:
            text = (await head.text(strip_padding=True, offset=self.current_variable_offset))
            self.program_variables.set(self.current_variable, text, scores=self.current_variable_scores[:-1], diff="", montonicity="fin")
            self.current_variable = None
            
            self.num_variables_decoded += 1
            # update hypothesis head state to reflect that we have finished decoding a variable
            self.last_state.num_variables_decoded = self.num_variables_decoded
            
            self.current_variable_scores = ()

            # extend prompt
            self.prompt += text
            old_prompt = self.prompt

            while self.current_variable is None and not self.query_head.done:
                await self.advance()

            if old_prompt != self.prompt or self.query_head.done:
                # print(f"detokenized decoder head IDs:\n", await head.detokenize(head.input_ids))
                # old_prompt_tokens = await head.tokenize(old_prompt)
                # new_prompt = await head.tokenize(self.prompt)
                appended_ids = await head.tokenize(self.prompt[len(old_prompt):])
                
                # if query is finished, add eos token to appended IDs (indicates to decoder that this sequence/query has finished decoding)
                if self.query_head.done: appended_ids += [head.eos_token_id]
                
                self.last_seen_input_ids = head.input_ids_without_padding.tolist() + appended_ids
                self.current_variable_offset = len(head.input_ids_without_padding) + len(appended_ids)

                return RewrittenInputIds(appended_input_ids=appended_ids, strip_eos=True)
            else:
                # set last_seen_input_ids to current input_ids w/o eos token
                self.last_seen_input_ids = head.input_ids_without_padding.tolist()
                self.current_variable_offset = len(head.input_ids_without_padding)

                return RewrittenInputIds(appended_input_ids=None, strip_eos=True)
        return None

    def matches(self, decoder_head: DecoderHead, allow_none_match=False):
        if self.last_seen_input_ids is None and allow_none_match:
            # this should only happen for the root hypothesis
            return True
        def tolist(t):
            if type(t) is list: return t
            else: return t.tolist()
        
        # make sure that the decoder head is at the same variable as the hypothesis head
        decoder_variable = decoder_head.user_data.get("head", {}).get("variable", None)
        if self.current_variable is None:
            if decoder_variable != "__done__" and decoder_variable is not None:
                return False
        else:
            variable_recurrence = self.recurring_variable_counter[self.current_variable]
            head_variable = f"{self.current_variable}[{variable_recurrence}]"
            if decoder_variable != head_variable:
                return False
        
        # make sure the input_ids align
        return str(list(reversed(tolist(self.last_seen_input_ids)))) == str(list(reversed(tolist(decoder_head.input_ids_without_padding))))

class HypothesesBasedHeadPool:
    def __init__(self, intepreter, root_head):
        self.intepreter = intepreter

        self.initial_prompt = None

        self.root_head = root_head
        
        # self.num_generations = -1
        # self.previous_generation = -1
        # self.previous_generation_heads = []
        self.heads = {}
        self.heads_done = []
        self.head_bucket_mapping = {}

        self.head_ctr = -1

    def inc_head_ctr(self):
        self.head_ctr += 1
        return self.head_ctr

    async def prepare(self):
        await self.root_head.advance()
        self.initial_prompt = self.root_head.prompt

    def replace_head(self, head_to_replace, head_to_replace_with):
        if head_to_replace == self.root_head: 
            self.root_head = head_to_replace_with
        else:
            num_tokens = len(head_to_replace.last_seen_input_ids)
            i = self.heads[num_tokens].index(head_to_replace)
            self.heads[num_tokens][i] = head_to_replace_with

    async def get_head(self, decoder_head: DecoderHead, readonly=False) -> HypothesisHead:
        # determine set of matching heads
        num_tokens = len(decoder_head.input_ids_without_padding)
        candidate_heads = self.heads.get(num_tokens, [])
        matching_heads = [h for h in candidate_heads if h.matches(decoder_head)]

        if len(matching_heads) == 0:
            matching_heads = [self.root_head]

        if not (len(matching_heads) > 0):
            for depth, h in [(depth, h) for depth in self.heads for h in self.heads[depth]] + [(0, self.root_head)]:
                    if h.last_seen_input_ids is not None:
                        print("Head with (depth {}, head_index {}, prototype {}) '{}'".format(depth, h.head_index, h.prototype, [await decoder_head.detokenizer_fn(h.last_seen_input_ids)]))
                    else:
                        print("Head with (depth {}, head_index {}, prototype {}) '{}'".format(depth, h.head_index, h.prototype, "None"))

        assert len(matching_heads) > 0, "fatal: (head_index {}) no matching HypothesisHead for depth {} and decoder sequence {}"\
            .format(decoder_head.seq_idx, num_tokens, [await decoder_head.text()])
        
        # sort such that prototypes are in the back
        matching_heads = sorted(matching_heads, key=lambda h: 0 if not h.prototype else 1)

        # for read-only interaction we do not branch (e.g. where clause validation)
        if readonly: 
            # last_seen = matching_heads[0].last_seen_input_ids
            # if last_seen is not None:
            #     last_seen = [await decoder_head.detokenize(last_seen)]
            # print("matching head", [await decoder_head.text()], last_seen)
            return matching_heads[0]

        # find predecessor head
        previous_gen_head: HypothesisHead = matching_heads[0]
        if previous_gen_head.prototype:
            # create new head from prototype
            head = previous_gen_head.copy()
            head.head_index = self.inc_head_ctr()
            head.prototype = False
            head.prototype_head_index = previous_gen_head.prototype_head_index
        else:
            # use predecessor head and keep a copy as prototype for additional branches
            head = previous_gen_head
            prototype_to_keep = head.copy()
            prototype_to_keep.prototype = True
            prototype_to_keep.prototype_head_index = head.head_index
            prototype_to_keep.head_index = self.inc_head_ctr()
            self.replace_head(head, prototype_to_keep)

        self.add_head(head)

        return head

    def add_head(self, head: HypothesisHead):
        num_tokens = len(head.last_seen_input_ids) if head.last_seen_input_ids is not None else 0
        if num_tokens not in self.heads:
            self.heads[num_tokens] = []
        self.heads[num_tokens].append(head)
        self.head_bucket_mapping[head.head_index] = num_tokens

    def update_head(self, head: HypothesisHead):
        old_num_tokens = self.head_bucket_mapping[head.head_index]
        num_tokens = len(head.last_seen_input_ids) if head.last_seen_input_ids is not None else 0
        
        # remove heads with query_head.done 
        if head.query_head.done:
            self.heads[old_num_tokens].remove(head)
            self.heads_done.append(head)
            return

        if old_num_tokens != num_tokens:
            self.heads[old_num_tokens].remove(head)
            self.add_head(head)

    async def advance_generation_if_needed(self, generation, detokenizer=None):
        pass # we keep all the heads
        
        # if self.previous_generation == generation:
        #     return
        # else:
        #     self.num_generations += 1
            
        #     self.previous_generation = generation
        #     # only keep .last_seen_input_ids == None heads around for the first generation
        #     self.previous_generation_heads = [h for h in self.heads if h.last_seen_input_ids is not None or self.num_generations == 0]
        #     self.heads = []

        #     if detokenizer is not None:
        #         print("Generation {} ({} heads)".format(self.num_generations, len(self.previous_generation_heads)))
        #         for h in self.previous_generation_heads:
        #             print("Head with (head_index {}, prototype {}) '{}'".format(h.head_index, h.prototype, [await detokenizer(h.last_seen_input_ids) if h.last_seen_input_ids is not None else "None"]))

    async def run(self):
        if self.initial_prompt is None:
            await self.prepare()

        self.rewrite_cycle = 0

        async def where(head: DecoderHead, return_user_data=False):
            # # determine generation from rewrite cycle
            # if head.seq_idx == 0: self.rewrite_cycle += 1
            # generation = math.floor(self.rewrite_cycle / 2)
            
            # await self.advance_generation_if_needed(generation)

            interpreter_head = await self.get_head(head, readonly=True)

            if head.is_at_eos():
                return head.next_token_logits
            logits = await interpreter_head.where_clause_logit_processor(head)

            if return_user_data:
                return logits, {"head": await interpreter_head.json()}
            
            return logits
        
        async def rewriter(decoder_head: DecoderHead): 
            head = await self.get_head(decoder_head)
            res = await head.head_input_id_rewriter(decoder_head)

            # update head storage location
            self.update_head(head)
            
            # keep track of the assigned interpreter head during further decoding
            head_data = await head.json()
            return RewrittenInputIds.with_user_data(res, "head", head_data)

        async def active_prompter(decoder_head): 
            # need to allow branching here because one where call might be followed by multiple active prompting calls 
            head = await self.get_head(decoder_head)
            return await head.active_prompt(decoder_head)

        query_result: Optional[List[DecoderHead]] = await self.intepreter.model.query(self.initial_prompt, where, rewriter, active_prompter)

        # collect query result
        head_result = []
        # if query() provides a set result heads use that
        if query_result is not None:
            for decoder_head in query_result:
                matching_done_interpreter_heads = [h for h in self.heads_done if h.matches(decoder_head)]
                if len(matching_done_interpreter_heads) > 0:
                    head_result += [h.query_head.result for h in matching_done_interpreter_heads]
                else:
                    print("warning: decoder designates head as done, but according to prompt interpreter it has not been completed yet:", await decoder_head.text(), "\n Did you specify a maximum length that has been exceeded?")
                    # print(decoder_head.user_data, self.heads)
        else:
            # otherwise use all heads that are .done
            for h in self.heads_done:
                if h.query_head.result is not None:
                    head_result.append(h.query_head.result)
        
        if len(head_result) == 0 and self.intepreter.distribution_variable is not None:
            print("warning: no heads were completed, but a distribution variable was specified. Cannot compute distribution.")

        if len(head_result) == 0:
            print("warning: the query has no valid result.")

        if len(head_result) == 1:
            return head_result[0]
        else:
            return head_result

class PromptInterpreter(LMQLRuntime):
    def __init__(self, force_model=None):
        self.program_variables = ProgramState(runtime=self)
        self.where = None
        self.model = None

        self.force_model = force_model

        self.specified_decoder_args = {}

        self.preview = True
        self.output_writer: DebuggerOutputWriter = PrintingDebuggerOutputWriter()

        self.distribution_variable = None
        self.distribution_values = None

        # default configurations
        self.configuration = {
            "follow_masking": True,
            "backtracking": False,
            "stats": False,
            "prefers_compact_mask": False
        }

    def on_update_configuration(self, key):
        if key == "follow_masking" and not self.configuration["follow_masking"]:
            print("info: follow masking is disabled, where clause will be ignored")
        if key == "backtracking" and self.configuration["backtracking"]:
            print("info: running with backtracking enabled")
        if key == "prefers_compact_mask":
            print("info: running with prefers_compact_mask = {}".format(self.configuration["prefers_compact_mask"]))
            self.prefers_compact_mask = self.configuration["prefers_compact_mask"]

    def print_stats(self):
        if not self.configuration["stats"]:
            return

        client = self.model.served_model

        print("Queries: {}".format(client.num_queries))
        print("Tokens/Prompt: {}".format(client.consumed_tokens))
        print("generate() calls: {}".format(client.num_generate_calls))
        print("billable tokens: {}".format(client.billable_tokens))

    async def run(self, fct, *args, **kwargs):
        # intercept symbol table entry for input
        if "input" in kwargs.keys() and kwargs["input"] == input:
            kwargs["input"] = self.input
        
        root_head = HypothesisHead(InterpretationHead(fct, None, args = args, kwargs=kwargs), self)
        pool = HypothesesBasedHeadPool(self, root_head)
        return await pool.run()

    def debugger_output(self, variable, head, prompt, where, trace, is_valid, is_final, mask, num_tokens, program_variables):
        if self.output_writer is None: return
        self.output_writer.add_interpreter_head_state(variable, head, prompt, where, trace, is_valid, is_final, mask, num_tokens, program_variables)

    def set_distribution(self, distribution_variable, values):
        self.distribution_variable = distribution_variable
        self.distribution_values = values

    async def input(self, *args):
        """Uses the output_writer input() implementation if available."""
        if hasattr(self.output_writer, "input"):
            return await self.output_writer.input(*args)
        else:
            return input(*args)

    def set_model(self, model):
        if self.force_model:
            # print("Forcing use of model {}, instead of {}.".format(self.force_model, model))
            model = self.force_model

        client = LMQLModelRegistry.get(model)

        # setup the VocabularyMatcher to use the concrete vocabulary of the model
        VocabularyMatcher.init(client.get_tokenizer())
        
        # for OpenAI models we optimize for compact logit masks
        if model.startswith("openai/"):
            self.prefers_compact_mask = True
            self.configuration["prefers_compact_mask"] = True

        self.model = client

    def set_decoder(self, method, **kwargs):
        assert self.model is not None, "Cannot set_decoder() before setting the model."
        self.specified_decoder_args = kwargs

        for key in self.configuration.keys():
            if key in kwargs:
                self.configuration[key] = kwargs[key]
                self.on_update_configuration(key)
        
        # pass output writer to model and decoder
        kwargs["output_writer"] = self.output_writer

        self.model.set_decoder(method, **kwargs)

    def set_where_clause(self, op):
        self.where = op


import sys
import ast
import types
import astunparse
import inspect
from lmql.language.fragment_parser import LMQLQuery
from lmql.language.compiler import PromptScope, SNFList, WhereClauseTransformation
from lmql.ops.ops import NextToken
from lmql.runtime.prompt_interpreter import ProgramState, digest

global show_transformed
show_transformed = False

SEQ = "<SEQ placeholder>"

class LMQLExpr: 
    def __init__(self, node):
        self.node = node
    
    def follow(self, seq, variable_name="SEQ"):
        for (result, final), follow_map in self.digest(seq, variable_name=variable_name, return_follow_map=True):
            yield follow_map

    def digest(self, seq, variable_name="SEQ", return_follow_map=False, debug=False):
        sequence = seq
        if type(sequence) is str:
            sequence = sequence.split(" ")
        full_text = ""

        program_variables = ProgramState()
        
        digested = []
        results = []

        for tok in sequence:
            added = (" " if len(full_text) > 0 else "") + tok
            full_text += added
            text = full_text

            # current context
            program_variables = program_variables.copy()
            program_variables.set(variable_name, text, "inc")

            # follow context
            follow_program_variables = program_variables.copy()
            follow_program_variables.set(variable_name, text + NextToken, "inc")

            # digest token with where expr
            result, is_final, trace = digest(self.node,
                context=program_variables,
                follow_context=follow_program_variables
            )

            if type(result) is list: 
                result = tuple(result)
            
            if return_follow_map:
                yield (result, is_final), self.node.follow_map
            else:
                if debug:
                    print(full_text, result, is_final)
                yield (result, is_final)

class LMQLExpressionCompiler(ast.NodeTransformer):
    def __init__(self, expr_compiler):
        self.expr_compiler: WhereClauseTransformation = expr_compiler
        self.expr_compiler.scope.defined_vars.add("SEQ")

    def visit_Assert(self, node):
        if node.msg is None:
            node.msg = ast.Constant(f"Line {node.lineno}: Assertion '{astunparse.unparse(node).strip()}' failed.", kind="str")
        return node
    
    def visit_Call(self, node):
        translated_calls = set(["inc", "dec", "fin", "var"])
        if type(node.func) is ast.Name and node.func.id in translated_calls:
            return ast.Call(
                ast.Name("raw_constant"),
                node.args + [ast.Constant(node.func.id, "str")],
                []
            )
        node.args = [self.visit(a) for a in node.args]

        return node

    def visit_Assign(self, node: ast.Assign):
        if type(node.value) is ast.Call:
            if type(node.value.func) is ast.Name and node.value.func.id == "LMQLExpr":
                call = self.visit(node.value)
                snf = SNFList()
                
                direct_result = self.expr_compiler.transform_node(call.args[0], snf)
                value = ast.Name(snf.last_var()) if snf.var_counter > 0 else ast.parse(direct_result.strip())

                return ast.copy_location(ast.Expr([
                    ast.parse("import lmql.runtime.lmql_runtime as lmql"),
                    snf.ast(),
                    ast.Assign(
                        node.targets, 
                        ast.Call(ast.Name("LMQLExpr"), [value], [])
                    )
                ]), node)
        
        return node

def lmql_test(fct):
    source = inspect.getsource(fct)
    startline = inspect.getsourcelines(fct)[1]
    source = '\n'.join(source.splitlines()[1:])
    source = " \n" * startline + source

    bogus_query = LMQLQuery()
    bogus_query.scope = PromptScope()
    bogus_query.scope.defined_vars = set()
    expr_compiler = WhereClauseTransformation(bogus_query)
    compiler = LMQLExpressionCompiler(expr_compiler)

    m = ast.parse(source)
    m = compiler.visit(m)

    code = astunparse.unparse(m)

    global show_transformed
    if show_transformed: print(code.strip())
    
    recompiled = compile(code, fct.__code__.co_filename, 'exec')
    return types.FunctionType(recompiled.co_consts[0], fct.__globals__)

def fin(v):
    return (v, "fin")

def var(v):
    return (v, "var")

def inc(v):
    return (v, "inc")

def dec(v):
    return (v, "dec")

def run_all_tests(g):
    g = g.copy()
    num_errors = 0
    for k in list(g.keys()):
        try:
            if k.startswith("test"): 
                g[k]()
        except AssertionError as e:
            print(e)
            num_errors += 1
    
    if num_errors != 0: 
        print(num_errors, "test(s) failed.")
        sys.exit(1)
    else: 
        print("All tests passed.")
        sys.exit(0)

def enable_show_transformed():
    global show_transformed
    show_transformed =True


from lmql.runtime.decoder_head import DecoderHead

class GroupByPostprocessor:
    def __init__(self):
        pass

    async def process(self, results, interpreter):
        raise NotImplementedError()


import os

openai_secret = None
openai_org = None

# get project root
ROOT_DIR = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

def get_openai_secret():
    if "LMQL_BROWSER" in os.environ:
        # get openai secret from JS context
        import js
        openai_secret = str(js.get_openai_secret())
        openai_org = str(js.get_openai_organization())
        
        return openai_secret, openai_org
    elif "LMQL_OPENAI_SECRET" in os.environ and "LMQL_OPENAI_ORG" in os.environ:
        return os.environ["LMQL_OPENAI_SECRET"], os.environ["LMQL_OPENAI_ORG"]
    elif "OPENAI_API_KEY" in os.environ:
        return os.environ["OPENAI_API_KEY"], ""
    else:
        search_paths = [
            os.path.join(ROOT_DIR, "api.env"),
            os.path.join(os.getcwd(), "api.env"),
            os.path.join(os.getenv("HOME"), ".lmql", "api.env")
        ]
        
        if not any(os.path.exists(p) for p in search_paths):
            m = """api.env not found in any of the following locations:\n\n{}\n\n To use OpenAI models you need to create an api.env file with the following contents:
        openai-secret: <your openai secret>
        openai-org: <your openai org>
        
        Alternatively, you may just define the environment variable OPENAI_API_KEY=sk-...
        """.format("\n".join(" - " + p for p in search_paths))
            raise FileNotFoundError(m)

        valid_paths = [p for p in search_paths if os.path.exists(p)]

        # get openai secret from file
        with open(valid_paths[0], "r") as f:
            for line in f:
                if line.startswith("openai-secret: "):
                    openai_secret = line.split("openai-secret: ")[1].strip()
                elif line.startswith("openai-org: "):
                    openai_org = line.split("openai-org: ")[1].strip()
                    
        return openai_secret, openai_org
                
openai_secret, openai_org = get_openai_secret()


from dataclasses import dataclass

from lmql.ops import *
from lmql.runtime.decoder_head import DecoderHead
from lmql.runtime.rewriter import RewrittenInputIds
from lmql.runtime.model_registry import LMQLModelRegistry
from lmql.runtime.program_state import ProgramState

import lmql.ops.token_set as token_set

from lmql.language.qstrings import qstring_to_stmts, TemplateVariable, DistributionVariable
from lmql.runtime.multi_head_interpretation import InterpretationHead, InterpreterCall, InterpreterCall, InterpreterHeadPool

import numpy as np
from lmql.utils import nputil

def make_backtracker_mask(subtoken_id):
    mask = np.zeros(VocabularyMatcher.instance().vocab_size, dtype=np.bool_)
    mask[subtoken_id] = True
    return mask

class BacktrackerHeadMixin:
    def add_backtracker_mask(self, index, token_id):
        # backtracking masks (used to prevent backtracking into the same branch)
        if index not in self.backtracker_masks:
            self.backtracker_masks[index] = make_backtracker_mask(token_id)
        else:
            self.backtracker_masks[index] = np.logical_or(self.backtracker_masks[index], make_backtracker_mask(token_id))

        # remove backtracker masks that are no longer needed
        for k in self.backtracker_masks.keys():
            if k > index: del self.backtracker_masks[k]

    def get_backtracker_mask(self, head: DecoderHead):
        index = len(head.input_ids_without_padding)
        return self.backtracker_masks.get(index, None)

    async def backtracking_check_valid(self, head: DecoderHead):
        where = self.interpreter.where
            
        text = (await head.text(self.current_variable_offset, strip_padding=True))
        diff_text = (await head.text(max(self.current_variable_offset, self.num_last_seen_ids - 1), strip_padding=True))

        # current context
        program_variables: ProgramState = self.program_variables.copy()
        program_variables.set(self.current_variable, text, scores=self.current_variable_scores, diff=diff_text, montonicity="inc")

        # evaluate where clause
        valid, is_final, _ = digest(where, context=program_variables, follow_context=None, no_follow=True)

        if not valid and is_final == "fin":
            self.add_backtracker_mask(head.input_ids_without_padding.size(0) - 1, make_backtracker_mask(head.input_ids_without_padding[-1]))
            # self.add_backtracker_mask(head.input_ids_without_padding.size(0), make_backtracker_mask(head.input_ids_without_padding[-1]))
            return False
        return True


class LMQLModelRegistry: pass

model_name_aliases = {
    "chatgpt": "openai/gpt-3.5-turbo",
    "gpt-4": "openai/gpt-4",
}

LMQLModelRegistry.registry = {}
# instance of model clients in this process
LMQLModelRegistry.clients = {}

def get_model(model):
    client = LMQLModelRegistry.clients.get(model, None)

    if client is None:
        # use auto connector to obtain model connection
        if LMQLModelRegistry.autoconnect is not None and model not in LMQLModelRegistry.registry:
            LMQLModelRegistry.autoconnect(model)

        client = LMQLModelRegistry.registry[model]()
        LMQLModelRegistry.clients[model] = client

    return client

LMQLModelRegistry.get = get_model
LMQLModelRegistry.autoconnect = None


import asyncio
from dataclasses import dataclass
from typing import Optional
import requests
import os
import torch
import time
import sys
import warnings
from tqdm import tqdm
from transformers import BeamScorer, LogitsProcessorList, PretrainedConfig, StoppingCriteriaList

from lmql.model.async_generation_utils import GenerationMixin

from lmql.utils import nputil
import numpy as np

class ServedModel:
    def __init__(self, host, model_identifier, timeout=60, use_tq=True, local=False):
        self.host = host
        
        self.model_identifier = model_identifier
        self.pending = {}
        self.polling_interval = 0.05
        self.timeout = timeout
        
        if local:
            from .local_client import LocalClient
            self.local_client = LocalClient(model_identifier)
        else:
            self.local_client = None

        # derive client_id from process id
        self.client_id = f"{os.getpid()}:{id(self)}"

        self.result_process_running = False

        if use_tq: 
            assert False
            self.tq = tqdm(desc="Querying model...", unit="tok")
        else: 
            self.tq = None

        self.sample_id = 0

        # track number of consumed tokens, queries (forward) and generate calls
        self.consumed_tokens = 0
        self.num_queries = 0
        self.num_generate_calls = 0
        self.billable_tokens = 0

    def report_stats(self, printer, decoder_step=None):
        if printer is None:
            print("no printer")
            return
        if hasattr(printer, "report_model_stats"):
            data = {
                "tokens": self.billable_tokens,
                "model": self.model_identifier
            }
            if decoder_step is not None:
                data["_step"] = decoder_step
            printer.report_model_stats(**data)

    async def generate(self, input_ids=None, *args, **kwargs):
        self.num_generate_calls += 1
        
        result = await super().generate(*args, input_ids=input_ids, **kwargs)
        
        if torch.is_tensor(result): ids = result
        else: ids = result["sequences"]
        self.billable_tokens += ids.numel()
        
        return result

    def create_result_processor_task_if_required(self):
        if self.result_process_running: 
            return
        
        loop = asyncio.get_event_loop()
        loop.create_task(self.result_processor())
        self.result_process_running = True

    def process_result(self, result):
        # check if torch is available
        if "next_token_logits" in result:
            result["next_token_logits"] = torch.tensor(result["next_token_logits"])
        if "input_ids" in result:
            result["input_ids"] = result["input_ids"]
        
        if "next_token_logits" in result:
            if result["next_token_logits"].dim() == 2:
                if self.tq is not None: 
                    self.tq.update(result["next_token_logits"].shape[0])
            else:
                if self.tq is not None: 
                    self.tq.update(1)

    async def result_processor(self):
        loop = asyncio.get_event_loop()

        while True:
            start_time = time.time()
            
            results = await loop.run_in_executor(None, requests.get, f"{self.host}/results/{self.client_id}")
            results = results.json()
            
            # process results
            for result in results:
                sample_id = result['sample_id']
                if sample_id in self.pending:
                    self.process_result(result)
                    self.pending[sample_id].set_result(result)
                    del self.pending[sample_id]
                else:
                    print("warning: result for unknown sample_id {}".format(sample_id), self.pending)

            remaining_sleep_time = self.polling_interval - (time.time() - start_time)
            if remaining_sleep_time > 0: await asyncio.sleep(remaining_sleep_time)

    def _timeout(self, sample_id):
        if sample_id in self.pending:
            self.pending[sample_id].set_exception(TimeoutError("Timed out waiting for result of sample {}".format(sample_id)))
            del self.pending[sample_id]

    async def detokenize(self, input_ids):
        self.create_result_processor_task_if_required()
        loop = asyncio.get_event_loop()

        # handle torch tensors
        if type(input_ids) is torch.Tensor:
            input_ids = input_ids.tolist()

        sample_id = self.sample_id
        self.sample_id += 1
        
        payload = {
            "action": "detokenize",
            "client_id": self.client_id,
            "input_ids": input_ids,
            "sample_id": sample_id
        }

        try:
            assert sample_id not in self.pending, "sample_id {} already in self.pending".format(sample_id)
            # setup future and timeout
            self.pending[sample_id] = loop.create_future()
            r = requests.post(f"{self.host}/queue", json=payload)

            if r.status_code != 200:
                raise Exception(f"Error posting to {self.host}/queue: {r.status_code}")
            
            loop.call_later(self.timeout, self._timeout, sample_id)
            
            return (await self.pending[sample_id])["text"]
        except requests.exceptions.ConnectionError as e:
            # check for connection refused
            if "Connection refused" in str(e):
                raise Exception(f"Error connecting to {self.host}/queue. Please make sure an instance of the LMQL inference API for this model is running. To start it, run `python -m serve <MODEL>`.")
            else:
                raise e

    async def tokenize(self, text):
        self.create_result_processor_task_if_required()
        loop = asyncio.get_event_loop()

        assert type(text) == str, "The provided text for tokenize() must be str."

        sample_id = self.sample_id
        self.sample_id += 1

        payload = {
            "action": "tokenize",
            "text": text,
            "client_id": self.client_id,
            "sample_id": sample_id
        }

        try:
            assert sample_id not in self.pending, "sample_id {} already in self.pending".format(sample_id)
            # setup future and timeout
            self.pending[sample_id] = loop.create_future()
            
            r = requests.post(f"{self.host}/queue", json=payload)
            
            if r.status_code != 200:
                raise Exception(f"Error posting to {self.host}/queue: {r.status_code}")
            
            loop.call_later(self.timeout, self._timeout, sample_id)
            
            return (await self.pending[sample_id])["input_ids"]
        except requests.exceptions.ConnectionError as e:
            # check for connection refused
            if "Connection refused" in str(e):
                raise Exception(f"Error connecting to {self.host}/queue. Please make sure an instance of the LMQL inference API for this model is running. To start it, run `python -m serve <MODEL>`.")
            else:
                raise e

    def reset_stats(self):
        self.consumed_tokens = 0
        self.num_queries = 0

    def forward(self, input_ids, attention_mask=None):
        assert input_ids.numel() > 0, "input_ids must not be empty"
        # keep stats
        num_input_tokens = sum(len(x) for x in input_ids)
        self.consumed_tokens = self.consumed_tokens + num_input_tokens
        
        assert input_ids.dtype == torch.long, "input_ids must be of dtype torch.long"
        # count prompt tokens + 1 generated output token per sequence
        
        # use local client if hosted in process
        if self.local_client is not None:
            return self.local_client.forward(input_ids, attention_mask)

        self.create_result_processor_task_if_required()

        loop = asyncio.get_event_loop()

        # # handle torch tensors
        if type(input_ids) is torch.Tensor:
            input_ids = input_ids.tolist()

        sample_id = self.sample_id
        self.sample_id += 1

        payload = {
            "action": "forward",
            "input_ids": input_ids,
            **{"attention_mask": attention_mask.tolist() if attention_mask is not None else None},
            "model_identifier": self.model_identifier,
            "client_id": self.client_id,
            "sample_id": sample_id
        }

        try:
            assert sample_id not in self.pending, "sample_id {} already in self.pending".format(sample_id)
            # setup future and timeout
            self.pending[sample_id] = loop.create_future()
            
            r = requests.post(f"{self.host}/queue", json=payload)
            
            if r.status_code != 200:
                raise Exception(f"Error posting to {self.host}/queue: {r.status_code}, {r.text}")
            
            loop.call_later(self.timeout, self._timeout, sample_id)
            
            return self.pending[sample_id]
        except requests.exceptions.ConnectionError as e:
            # check for connection refused
            if "Connection refused" in str(e):
                raise Exception(f"Error connecting to {self.host}/queue. Please make sure an instance of the LMQL inference API for this model is running. To start it, run `python -m serve <MODEL>`.")
            else:
                raise e


class Sample:
    def __init__(self, model, seq=None, input_ids=None):
        assert not (self.seq is None and self.input_ids), "Either seq or input_ids must be provided for a Sample()"
        
        self.model = model
        self._seq = seq
        self._input_ids = input_ids
    
    async def seq(self):
        if self._seq is None:
            assert self._input_ids is not None
            self._seq = await self.model.detokenize(self._input_ids)
        return self._seq
    
    async def input_ids(self):
        if self._input_ids is None:
            assert self.seq is not None
            self._input_ids = await self.model.tokenize(self._seq)
        return self._input_ids

    async def successor(self, token_id) -> 'Sample':
        return Sample(self.model, input_ids=(await self.input_ids()) + [token_id])

    async def distribution(self):
        res = await self.model.forward(await self.input_ids())
        return res["next_token_logits"]
    
    def __repr__(self) -> str:
        return "<Sample seq={} input_ids={}>".format(self._seq, self._input_ids)

    async def concat(self, text):
        seq = (await self.seq()) + text
        input_ids = await self.model.tokenize(seq)
        
        return Sample(self.model, seq=seq, input_ids=input_ids)

@dataclass
class ServedPretrainedModelOutput:
    logits: torch.Tensor

    # enable 'in' operator
    def __contains__(self, item):
        return item in self.__dict__.keys()

class ServedPretrainedModel(ServedModel, GenerationMixin):
    def __init__(self, host, model_identifier, timeout=60, use_tq=False, model_kwargs=None, local=False):
        super().__init__(host, model_identifier, timeout, use_tq=use_tq, local=local)

        self.model_kwargs = model_kwargs or {}
        
        self.config = PretrainedConfig()
        self.device = torch.device("cpu")

        # old context model
        self.previous_context_model = None

    async def __call__(self, input_ids: torch.Tensor, *args, **kwargs):
        res = await self.forward(input_ids, attention_mask=kwargs.get("attention_mask", None))
        logits = res["next_token_logits"]

        return ServedPretrainedModelOutput(logits.unsqueeze(1))

    async def __aenter__(self):
        self.previous_context_model = ServedPretrainedModel.context_model

        bos_token_id = (await self.tokenize("<BOS>"))[0]
        setattr(self, "bos_token_id", bos_token_id)
        eos_token_id = (await self.tokenize("<EOS>"))[0]
        setattr(self, "eos_token_id", eos_token_id)

        ServedPretrainedModel.context_model = self
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        ServedPretrainedModel.context_model = self.previous_context_model
        self.previous_context_model = None

ServedPretrainedModel.context_model = None


async def main():
    model = ServedPretrainedModel("http://localhost:8080", "gpt2")

    bos_token_id = (await model.tokenize("<BOS>"))[0]
    eos_token_id = (await model.tokenize("<EOS>"))[0]

    print([await model.detokenize([int(v) if not v.endswith(",") else int(v[:-1]) for v in sys.argv[1:]])])

if __name__ == "__main__":
    asyncio.run(main())



from typing import List, Any
import asyncio
from typing import Optional
from dataclasses import dataclass
from abc import abstractclassmethod, abstractmethod

import numpy as np
from lmql.utils import nputil

@dataclass
class Continuation:
    token: Any
    logprob: Any
    user_data: Any

class criterion: 
    def __and__(self, other):
        return logical_and(self, other)

    def __or__(self, other):
        return logical_or(self, other)

    def __not__(self):
        return logical_not(self)

class logical_not(criterion):
    def __init__(self, a):
        self.a = a
    def __call__(self, seq):
        return not self.a(seq)

class logical_and(criterion):
    def __init__(self, a, b):
        self.a = a
        self.b = b
    def __call__(self, seq):
        return self.a(seq) and self.b(seq)

class logical_or(criterion):
    def __init__(self, a, b):
        self.a = a
        self.b = b
    def __call__(self, seq):
        return self.a(seq) or self.b(seq)


class eos_criterion(criterion): 
    def __call__(self, seq):
        return seq.is_done()
eos = eos_criterion()

class ge(criterion):
    def __init__(self, max_len):
        self.max_len = max_len
    def __call__(self, seq):
        return len(seq) >= self.max_len

class lt(criterion): 
    def __init__(self, max_len):
        self.max_len = max_len
    def __call__(self, seq):
        return len(seq) < self.max_len


class topk_scorer:
    @abstractmethod
    def __call__(self, score, *args, **kwargs):
        pass

    @classmethod
    @abstractmethod
    def score(cls, scores, *args, **kwargs):
        pass


class sum_scorer(topk_scorer):
    def __call__(self, scores, s, *args, **kwargs):
        return self.score(scores, s, *args, **kwargs)

    @classmethod
    def score(cls, scores, s, *args, **kwargs):
        scores = scores[s.prompt_len:]
        if s.data("head.variable") == "__done__" and hasattr(s, "next_ids") and len(s.next_ids) == 0:
            scores = scores[:-1]
        return scores.sum()


class mean_scorer(topk_scorer):
    def __call__(self, scores, s, *args, **kwargs):
        return self.score(scores, *args, **kwargs)

    @classmethod
    def score(cls, scores, s, *args, **kwargs):
        scores = scores[s.prompt_len:]
        if s.data("head.variable") == "__done__" and hasattr(s, "next_ids") and len(s.next_ids) == 0:
            scores = scores[:-1]
        return scores.mean()


class alpha_length_normalized(topk_scorer):
    def __init__(self, alpha=0.7):
        self.alpha = alpha

    def __call__(self, scores, *args, **kwargs):
        return self.score(scores, self.alpha)

    @classmethod
    def score(cls, scores, alpha=0.7, *args, **kwargs):
        if len(scores) == 0:
            return np.array(0.0)
        else:
            return (np.array(1.0) / float(len(scores)) ** alpha) * scores.sum()


class alpha_length_normalized_det(topk_scorer):
    def __init__(self, alpha=0.7):
        self.alpha = alpha

    def __call__(self, scores, s, *args, **kwargs):
        return self.score(scores, s, alpha=self.alpha)

    @classmethod
    def score(cls, scores, s, alpha=0.7, *args, **kwargs):
        scores = scores[s.prompt_len:]
        det = s.deterministic[s.prompt_len:]
        if s.data("head.variable") == "__done__" and hasattr(s, "next_ids") and len(s.next_ids) == 0:
            scores = scores[:-1]
            det = det[:-1]
        score_det = alpha_length_normalized.score(scores[det], alpha)
        score_act = alpha_length_normalized.score(scores[~det], alpha)
        return score_det + score_act


def max_score(ar, scorer: Optional[topk_scorer]=None):
    """
    Returns the value of the maximum scoring sequence in ar.
    """
    assert len(ar) > 0, "dc.max_score is not defined for empty arrays"
    
    if scorer is None:
        scorer = alpha_length_normalized()

    def op_max_score(seqs, score):
        res = np.array([score] + [scorer(s.logprobs, s) for s in seqs])
        return res.max()

    return ar.reduce(op_max_score, np.finfo(np.float32).min)

def min_score(ar, scorer: Optional[topk_scorer]=None):
    """
    Returns the value of the minimum scoring sequence in ar.
    """
    assert len(ar) > 0, "dc.min_score is not defined for empty arrays"

    if scorer is None:
        scorer = alpha_length_normalized()

    def op_max_score(seqs, score):
        return np.array([score] + [scorer(s.logprobs, s) for s in seqs]).min()

    return ar.reduce(op_max_score, np.finfo(np.float32).max)

def topk(ar, k, scorer: Optional[topk_scorer]=None, name=None):
    if scorer is None:
        scorer = alpha_length_normalized()
    
    def op_topk(seqs):
        if len(seqs) == 0: 
            return None

        scores = np.stack([scorer(s.logprobs, s) for s in seqs], axis=0)
        _, topk_indices = nputil.topk(scores, min(k, len(scores)))

        return [seqs[i] for i in topk_indices]

    return ar.element_wise(op_topk, name=name)


def seperate_topk(ar, k, scorer: Optional[topk_scorer]=None, name=None):
    if scorer is None:
        scorer = alpha_length_normalized()
    
    def mark_topk(seqs):
        if len(seqs) == 0: 
            return None

        scores = np.stack([scorer(s.logprobs, s) for s in seqs], axis=0)
        _, topk_indices = nputil.topk(scores, min(k, len(scores)))
        for i in range(len(seqs)):
            seqs[i].data("_is_topk", i in topk_indices)
        return seqs

    def in_top_k(seqs):
        return [s for s in seqs if s.data("_is_topk")]

    def not_in_top_k(seqs):
        return [s for s in seqs if not s.data("_is_topk")]

    ar = ar.element_wise(mark_topk, name=name)
    ar_in_top_k = ar.element_wise(in_top_k, name=name)
    ar_not_in_top_k = ar.element_wise(not_in_top_k, name=name)
    return ar_in_top_k, ar_not_in_top_k

class ComponentWiseValue:
    def __init__(self, value):
        self.value = value

def componentwise_arg(arg, key):
        if type(arg) is ComponentWiseValue:
            assert key in arg.value, f"key {key} not found in component-wise operand"
            return arg.value[key]
        return arg
    
def componentwise_kwargs(key, **kwargs):
    return {k: componentwise_arg(v, key) for k, v in kwargs.items()}

def assert_shape_match(a, b, op, allow_mismatch_keys=False):
    assert type(a) is dict, "internal error: can only perform shape checks on dicts of lists, but got a {}".format(type(a))
    
    if type(b) is not dict:
        raise ValueError(f"cannot apply {op} to a grid-shaped sequence pool and a non-grid-shaped sequence pool")
    
    if allow_mismatch_keys:
        return True

    missing_keys = []

    for k in set(a.keys()):
        if k not in b:
            missing_keys.append(k)
    
    if len(missing_keys) > 0:
        raise ValueError(f"cannot apply {op} to two grid-shaped sequence pools with different sets of keys {a.keys()} vs. {b.keys()} (dimensions in a but not b: {missing_keys})")

    return True

class NoDim: pass

def apply_componentwise(op, a, b, opname, allow_mismatch_keys=False, default_value=None):
    assert_shape_match(a, b, opname, allow_mismatch_keys=allow_mismatch_keys)

    pools_a = dict(a.items() if type(a) is dict else [(NoDim, a)])
    pools_b = dict(b.items() if type(b) is dict else [(NoDim, b)])

    all_keys = set(pools_a.keys()).union(set(pools_b.keys()))
    result = {}

    for k in all_keys:
        pool_a = pools_a[k] if k in pools_a else default_value
        pool_b = pools_b[k] if k in pools_b else default_value

        result[k] = op(pool_a, pool_b)

    return result

class DataArray:
    def __init__(self, data, dims=None):
        super().__init__()
        
        if type(data) is not dict: data = {NoDim: data}

        try:
            self.dtype = type(list(data.values())[0][0]) if type(data) is dict else type(data)
        except:
            self.dtype = 'seq'

        self.sequences = data
        self.shape = dims

    @property
    def paths(self):
        return self.sequences.keys()

    def reduce(self, op, accumulated=None, *args, **kwargs):
        for element in self.sequences.values():
            if element is None: continue
            accumulated = op(element, accumulated, *args, **kwargs)
        return accumulated
    
    def name(self, name=None, nopath=False):
        """
        Names the sequences in this pool with the given name. 
        This information is useful for debugging purposes.

        Parameters:
            name (str): The name to give to the sequences in this pool.
            nopath (bool): If True, the name will not be suffixed with the path of the sequence in the pool (e.g. <name>.dim1.dim2).
        """
        if name is None:
            return self
        
        for path, result in self.sequences.items():
            if result is None:
                continue
            for s in result:
                if hasattr(s, "pool"):
                    if path is NoDim or nopath:
                        s.pool = name
                    else:
                        s.pool = name + "." + ".".join(path)
        return self

    def element_wise(self, op, name=None, *args, **kwargs):
        def op_with_path(path, element, *args, **kwargs):
            return path, op(element, *args, **kwargs)
        result_items = [op_with_path(path, element, *args, **kwargs) for path, element in self.sequences.items()]
        return DataArray(dict([(path,value) for path,value in result_items if value is not None]), dims=self.shape).name(name)

    async def aelement_wise(self, op, *args, **kwargs):
        async def op_with_path(path, element, *args, **kwargs):
            return path, await op(element, *args, **kwargs)
        result_items = await asyncio.gather(*[op_with_path(path, seqs, *args, **kwargs) for path, seqs in self.sequences.items()])
        return DataArray(dict(result_items))

    def extend(self, other):
        def op_extend(p1, p2):
            extended_seqs = []
            for sq, continuation in zip(p1, p2):
                tokens = continuation.token.reshape(-1)
                logprobs = continuation.logprob.reshape(-1)
                user_data = continuation.user_data or [None] * len(tokens)
                for t,s,u in zip(tokens, logprobs, user_data):
                    extended_seqs.append(sq.extend(Continuation(t, s, u)))
            return extended_seqs

        return DataArray(apply_componentwise(op_extend, self.sequences, other.sequences, "extend", allow_mismatch_keys=False), dims=self.shape)

    def separate_by(self, *criteria):
        """
        Separate this bundle intot two bundles (a,b), one containing sequences that satisfy 
        the provided criteria (a), and one containing sequences that do not (b).
        """
        def is_a(seq): return all(c(seq) for c in criteria)
        
        def op_separate(s):
            if s is None: return None, None
            return [s for s in s if is_a(s)], [s for s in s if not is_a(s)]
        tuple_array = self.element_wise(op_separate)

        return tuple_array.element_wise(lambda x: x[0]), tuple_array.element_wise(lambda x: x[1])

    def separate_by_list(self, criterium, reshape=False):
        """
        Separate this bundle into as many bundles as there are unique evaluations of criterium.
        """

        separated_seqs = {}

        for s in self.unstructured():
            key = criterium(s)
            if key in separated_seqs:
                separated_seqs[key].append(s)
            else:
                separated_seqs[key] = [s]

        for key, seqs in separated_seqs.items():
            if reshape:
                separated_seqs[key] = DataArray(separated_seqs[key]).reshape(self.shape)
            else:
                separated_seqs[key] = DataArray(separated_seqs[key])

        return separated_seqs

    def __add__(self, other):
        def op_add(s1, s2):
            if s1 is None: return s2
            if s2 is None: return s1
            return [s for s in (s1 + s2) if s is not None]

        result = apply_componentwise(op_add, self.sequences, other.sequences, "+", allow_mismatch_keys=True)
        return DataArray(result)

    def __len__(self):
        return sum(len(p) if p is not None else 0 for p in self.sequences.values())

    def unstructured(self):
        """
        Unstructured iterator function over all sequences in this array.
        """
        for leaf in self.sequences.values():
            if type(leaf) is list:
                for item in leaf:
                    if item is None:
                        continue
                    yield item
            else:
                if leaf is None:
                    continue
                yield leaf

    def flatten(self):
        """
        Returns a flattened DataArray listing all elements of this array in the same dimension.
        """
        return DataArray([s for s in self.unstructured()])

    def item(self, i, path=None):
        """
        Returns item `i` at the given path in this array.
        """
        if path is None:
            assert list(self.sequences.keys())[0] is NoDim, "Cannot access item without path if the array has more than one dimension."
            path = NoDim
        return self.sequences[path][i]

    def items(self, path=None):
        if path is None:
            assert list(self.sequences.keys())[0] is NoDim, "Cannot access item without path if the array has more than one dimension."
            path = NoDim
        for s in self.sequences[path]:
            yield s

    def unique(self):
        """
        Returns a new DataArray with the same shape as this one, but with only unique sequences within each pool.
        """
        return self.element_wise(lambda seqs: list(set(seqs)))

    def filter(self, op):
        """
        Filters the sequences in this array dimension-wise by the given operation.
        """
        def op_filter(seqs):
            r = [s for s in seqs if op(s)]
            if len(r) == 0:
                return None
            else:
                return r
        return self.element_wise(op_filter)

    def reshape(self, *dims):
        """
        Regroup all sequences in this array according to the provided list of dimensions.

        A dimension can be either a string or a callable.
        
        For strings it acts as a getter for the corresponding user data attribute 
        of each sequence (e.g. head.variable).

        For callables it is called with the sequence as argument and should return 
        a unique value by which the sequences are grouped (e.g. lambda s: len(s)).
        """
        if dims is None or (len(dims) == 1 and dims[0] is None):
            return self

        # unpack if necessary
        if len(dims) == 1 and (type(dims[0]) is list or type(dims[0]) is tuple):
            dims = dims[0]
        
        dims_computer = [d if callable(d) else lambda s: s.data(d) for d in dims]

        seqs = [s for s in self.unstructured()]
        dimensions = [tuple(d(s) for d in dims_computer) for s in seqs]
        data = {}
        
        for s,d in zip(seqs, dimensions):
            if d in data: data[d].append(s)
            else: data[d] = [s]

        return DataArray(data, dims=dims)

    async def text(self, sequences=None):
        """Async __str__ which additionally detokenizes sequences for better readability."""
        if sequences is None: 
            sequences = self.sequences
        
        lines = []
        def str_k(k):
            if k is NoDim:  return "NoDim"
            elif k is None: return "None"
            else: return str(k)

        max_len_key = max([len(str_k(k)) for k in sequences.keys()])
        for k in sorted(sequences.keys()):
            dimension_lines = [await s.text() for s in self.sequences[k]]
            for i,l in enumerate(dimension_lines):
                if i == 0:
                    lines.append(str_k(k) + (max_len_key - len(str_k(k)) + 1) * " " + l)
                else:
                    lines.append(" " * max_len_key + " " + l)
        return "\n".join(lines)
    
    async def str(self, sequences=None):
        """Async __str__ which additionally detokenizes sequences for better readability."""
        if sequences is None: 
            sequences = self.sequences
        
        lines = []
        def str_k(k):
            if k is NoDim:  return "NoDim"
            elif k is None: return "None"
            else: return str(k)

        max_len_key = max([len(str_k(k)) for k in sequences.keys()])
        for k in sorted(sequences.keys()):
            dimension_lines = [await s.str() for s in self.sequences[k]]
            for i,l in enumerate(dimension_lines):
                if i == 0:
                    lines.append(str_k(k) + (max_len_key - len(str_k(k)) + 1) * " " + l)
                else:
                    lines.append(" " * max_len_key + " " + l)
        return "\n".join(lines)
    
    def __str__(self, data=None):
        if data is None: data = self.sequences

        if type(data) is list:
            lines = []
            for s in data:
                lines.append(str(s))
            return "\n".join(lines)

        lines = []
        def str_k(k):
            if k is NoDim:  return "NoDim"
            elif k is None: return "None"
            else: return str(k)

        if len(self) == 0:
            return "<empty>"

        paths = [len("\t".join(str_k(k))) for k in data.keys()]
        if len(paths) > 0:
            max_len_path = max([len(str_k(k)) for k in data.keys()])
            for k in sorted(data.keys(), key=lambda k: str_k(k)):
                dimension_lines = self.__str__(data[k]).split("\n")
                dimension_values = str_k(k)
                for i,l in enumerate(dimension_lines):
                    if i == 0:
                        lines.append(f"{dimension_values:{max_len_path}} {l}")
                    else:
                        lines.append(f"{'':{max_len_path}} {l}")
        else:
            lines.append("<empty>")
        return "\n".join(lines)

    def data(self, key=None, value=None, sticky=False):
        def op_data(sqs):
            return [s.data(key, value, sticky=sticky) for s in sqs]
        return self.element_wise(op_data)

def items_hierarchy(d, prefix=[]):
    items = []
    for k,v in d.items():
        k = str(k)
        if type(v) is dict:
            items += items_hierarchy(v, [*prefix, k])
        else:
            items.append((".".join([*prefix, k]), v))
    if len(prefix) == 0:
        return sorted(items, key=lambda x: x[0])
    return items

def seqs(seqs=None):
    return DataArray(seqs)



import sys
import termcolor

def error_loc_visualisation(file, loc, msg, color):
    start, end = loc

    print("")
    with open(file, "r") as f:
        for i, l in enumerate(f):
            if start.line - 1 == i:
                if start.line != end.line: end.column = len(l)
                print(str(i+1) + ":", l.rstrip())
                print(" " * (start.column + 2) + termcolor.colored("-" * (end.column - start.column + 1), color))
                print(" " * (start.column + 2), termcolor.colored(msg, color))
                break
            # elif start.line - 2 < i and start.line + 2 > i:
                # print(i+1, l.rstrip())
    print("")

def error(msg, node):
    error_loc_visualisation(node.file, node.loc, "error: " + msg, "red")
    sys.exit(1)



import sys
from lmql.ops.token_set import TokenSetSymbolic, TokenSetConcrete

import numpy as np
from lmql.utils import nputil

class DecoderHead:
    def __init__(self, head: int, input_ids, next_token_id, next_token_logits: np.ndarray, tokenizer_fn, detokenizer_fn, eos_token_id, initial_prompt_offset, next_token_logprob=None, user_data=None):
        self.seq_idx: int = head
        
        self.input_ids = input_ids
        
        self.next_token_id = next_token_id
        if nputil.is_array(self.next_token_id): 
            self.next_token_id = nputil.item(self.next_token_id)
        self.next_token_logits = next_token_logits
        
        self.tokenizer_fn = tokenizer_fn
        self.detokenizer_fn = detokenizer_fn

        self.eos_token_id = eos_token_id
        self.initial_prompt_offset = initial_prompt_offset

        self.next_token_logprob = next_token_logprob
        
        self.user_data = user_data

    def is_at_eos(self):
        return self.next_token_id == self.eos_token_id

    @property
    def input_ids_without_padding(self):
        t = self.input_ids
        while len(t) > 0 and t[0] == self.eos_token_id:
            t = t[1:]
        return t

    async def text(self, offset=None, strip_padding=False, upper_bound=None, strip_eos=False, return_ids=False):
        if offset is None:
            offset = 0
        if upper_bound is None:
            upper_bound = len(self.input_ids)
        
        if offset == 0:
            while self.input_ids[offset] == self.eos_token_id:
                offset += 1
        else:
            for i in range(len(self.input_ids)):
                if self.input_ids[i] != self.eos_token_id: break
                else: offset += 1
        ids = self.input_ids[offset:upper_bound]
        while strip_eos and ids[-1] == self.eos_token_id:
            ids = ids[:-1]
        
        if return_ids:
            return await self.detokenizer_fn(ids), ids
        else:
            return await self.detokenizer_fn(ids)

    async def tokenize(self, t):
        return await self.tokenizer_fn(t)

    async def detokenize(self, input_ids):
        return await self.detokenizer_fn(input_ids)

    async def translate_mask(self, mask, vocab_size):
        m = np.ones([vocab_size], dtype=np.int32)

        if mask == "*":
            return None
        elif type(mask) is TokenSetSymbolic:
            # start with full mask for minussets (e.g. * \ {a,b,c})
            if not mask.minusset: 
                m[:] = 0
            
            for t in mask.tokens:
                # switch to something more unique
                if t == "eos": t = "<EOS>"
                
                token_id = (await self.tokenize(t))[0]
                m[token_id] = 0 if mask.minusset else 1

            return m
        else:
            assert type(mask) is TokenSetConcrete
            return mask.mask
    
    async def make_mask(self, *token_ids, vocab_size):
        m = np.zeros([vocab_size])
        for t in token_ids: m[t] = 1
        return m


import torch
from transformers import AutoModelForCausalLM
import asyncio

class LocalClient:
    def __init__(self, model_identifier):
        self.model_identifier = model_identifier
        
        print("Using local in-process model", model_identifier)
        self.model = AutoModelForCausalLM.from_pretrained(model_identifier, device_map="auto")
        self.forward_queue = asyncio.Queue()
        self.forward_worker = asyncio.create_task(self.forward_worker_loop())
        print("Ready.")
    
    async def forward_worker_loop(self):
        while True:
            input_ids, attention_mask, fut = await self.forward_queue.get()
            
            res = self.model.forward(input_ids=input_ids, attention_mask=attention_mask)
            
            if input_ids.ndimension() == 2:
                next_token_logits = res.logits[:,-1]
            else:
                next_token_logits = res.logits[-1]
            
            fut.set_result({"next_token_logits": next_token_logits.detach()})
    
    def forward(self, input_ids, attention_mask):
        loop = asyncio.get_event_loop()
        fut = loop.create_future()
        self.forward_queue.put_nowait((input_ids, attention_mask, fut))
        return fut
    
    def __del__(self):
        self.forward_worker.cancel()